/*************************************************************************************
* The Petri-Net Add-on is (c) 2012 Plant Bioinformatics Group, IPK Gatersleben,
* http://bioinformatics.ipk-gatersleben.de
* The source code for this project, is available under the GPL license v2.0 
* available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html. By using this
* Add-on and VANTED you need to accept the terms and conditions of this
* license. For further details see GPL_license.txt in the root folder of this project.
************************************************************************************/

package petri_jfern;

import java.awt.Color;
import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import org.AttributeHelper;
import org.ReleaseInfo;
import org.rakiura.cpn.BasicNet;
import org.rakiura.cpn.InputArc;
import org.rakiura.cpn.Multiset;
import org.rakiura.cpn.Net;
import org.rakiura.cpn.OutputArc;
import org.rakiura.cpn.event.TransitionEvent;
import org.rakiura.cpn.event.TransitionFinishedEvent;
import org.rakiura.cpn.event.TransitionListener;
import org.rakiura.cpn.event.TransitionStartedEvent;
import org.rakiura.cpn.event.TransitionStateChangedEvent;
import org.rakiura.cpn.gui.NetViewer;

import simulation.Simulator;
import simulation.SimulatorListener;
import vanted.attribute.token.TokenAttributeContinuous;
import vanted.petrinetelements.AbstractArc;
import vanted.petrinetelements.AbstractToken;
import vanted.petrinetelements.InArc;
import vanted.petrinetelements.Library;
import vanted.petrinetelements.OutArc;
import vanted.petrinetelements.PetriNet;
import vanted.petrinetelements.PetriNetListener;
import vanted.petrinetelements.Place;
import vanted.petrinetelements.TokenContinuous;
import vanted.petrinetelements.TokenDiscrete;
import vanted.petrinetelements.Transition;
import vanted.petrinetelements.cpn.Marking;
import vanted.petrinetelements.cpn.TokenObject;
import vanted.petrinetelements.misc.PetriNetType;
import de.ipk_gatersleben.ag_nw.graffiti.GraphHelper;
import de.ipk_gatersleben.ag_nw.graffiti.services.task.BackgroundTaskHelper;
import de.ipk_gatersleben.ag_nw.graffiti.services.task.BackgroundTaskStatusProviderSupportingExternalCallImpl;

// TODO: Auto-generated Javadoc
/**
 * The Class PNJFern parses a VANTED Petri net into a rakiura Petri net, performs all simulation operations and writes the marking for each simulation step into a file (if enabled).
 *
 * @author Kevin Pucknat
 */
public class PNJFern implements TransitionListener, PetriNetListener, SimulatorListener, vanted.petrinetelements.TransitionListener, Library {
	
	/** The net. */
	private Net net;
	
	/** The place map. */
	private PlaceMap placeMap;
	
	/** The transition map. */
	private TransitionMap transitionMap;
	
	/** The arc map. */
	private ArcMap arcMap;
	
	/** The simlator */
	private Simulator sim = new Simulator();
	
	/** The Petri net. */
	private PetriNet pn;
	
	/** Shows token flow. */
	private boolean show = true;
	
	/** The highlight enabled transitions delay. */
	private int highlightEnabledTransitionsDelay = 0;
	
	private List<Transition> excludedTransitions = new ArrayList<Transition>();	//new
	
	/**
	 * Parses the VANTED Petri net into a rakiura Petri net.
	 *
	 * @param pn the pn (vanted)
	 * @return the net (rakiura)
	 */
	public Net parsePN2JFERN(PetriNet pn) {		
		this.pn = pn;
		
		for(Transition transition : pn.createRandomReversibleTransitions(pn.getAllTransitions()))		//new
			excludedTransitions.add(transition.getReversibleTransition());										//new
				
		pathExeAddons = ReleaseInfo.getAppSubdirFolderWithFinalSep("addons", "PetriNet");
		parseNet(pn);
		pn.addListener(this);

		return net;
	}
	
	/**
	 * Parses the net.
	 *
	 * @param pn the pn (vanted)
	 * @return the net (rakiura)
	 */
	private synchronized Net parseNet(PetriNet pn) {
		if (pn == null)
			return null;
		show = true;
		placeMap = new PlaceMap();
		transitionMap = new TransitionMap();
		arcMap = new ArcMap();
		
		net = buildBasicNet(pn);
		
		addArcWeightAndCapacity(pn);

//		 NetViewer viewer = new NetViewer(net);
//		 viewer.getFrame().show();
	
		return net;
	}
	
	/**
	 * Builds the structure of the rakiura net: transitions, places, arcs and tokens.
	 *
	 * @param pn the pn (vanted)
	 * @return the net (rakiura)
	 */
	private Net buildBasicNet(PetriNet pn) {
		BasicNet net = new BasicNet();
		
		for (AbstractArc arc : pn.getAllArcs()) {
			org.rakiura.cpn.Transition t;
			if (transitionMap.containsKeyV(arc.getTransition())) {
				t = transitionMap.getTransitionR(arc.getTransition());
			} else {
				t = new org.rakiura.cpn.Transition();
				net.add(t);
				transitionMap.addTransition(arc.getTransition(), t);
				t.setName(AttributeHelper.getLabel(arc.getTransition().getNode(), ""));
				arc.getTransition().addTransitionListener(this);
				t.addTransitionListener(this);
			}
			
			org.rakiura.cpn.Place p;
			if (placeMap.containsKeyV(arc.getPlace())) {
				p = placeMap.getPlaceR(arc.getPlace());
			} else {
				p = new org.rakiura.cpn.Place();
				net.add(p);
				placeMap.addPlace(arc.getPlace(), p);
				p.setName(AttributeHelper.getLabel(arc.getPlace().getNode(), ""));
				
				// add tokens
				if (arc.getPlace().getTokens().size() > 0) {
					p.addTokens(arc.getPlace().getTokens());
				}
			}
			
			if (arc instanceof InArc) {
				InArc inArc = (InArc) arc;
				InputArc inputArc = new InputArc(p, t);
				net.add(inputArc);
				arcMap.addArc(inArc, inputArc);
				
				// add InputArc Guards
				// addInputGuard(inArc, inputArc);
				
				// add simple input arc expression (like an arc weight)
				// addInArcExpressions(inArc, inputArc);
			} else
				if (arc instanceof OutArc) {
					OutArc outArc = (OutArc) arc;
					OutputArc outputArc = new OutputArc(t, p);
					net.add(outputArc);
					arcMap.addArc(outArc, outputArc);
					
					// add arc expression
					// addOutArcExpressions(outArc, outputArc);
				}
		}
		
		if (placeMap.getAllPlacesV().size() != pn.getAllPlaces().size()) {
			for (Place place : pn.getAllPlaces()) {
				if (!placeMap.getAllPlacesV().contains(place)) {
					org.rakiura.cpn.Place p = new org.rakiura.cpn.Place();
					net.add(p);
					placeMap.addPlace(place, p);
					
					// add tokens
					if (place.getTokens().size() > 0) {
						p.addTokens(place.getTokens());
					}
				}				
			}
		}
		if (transitionMap.getAllTransitionsV().size() != pn.getAllTransitions().size())
			for (Transition transition : pn.getAllTransitions()) {
				if (!transitionMap.getAllTransitionsV().contains(transition)) {
					org.rakiura.cpn.Transition t = new org.rakiura.cpn.Transition();
					net.add(t);
					transitionMap.addTransition(transition, t);
					transition.addTransitionListener(this);
					t.addTransitionListener(this);
				}
			}
		
		return net;
	}
	
	/** Steps to simulate. */
	private int step = 1;
	
	/** Write to file. */
	private boolean write2file = true;
	
	/** Writer. */
	private Writer writer;
	
	/**
	 * Write new marking from rakiura Petri net to VANTED Petri net.
	 *
	 * @param finished, false writes only token to changed places; true writes token to all places
	 */
	public synchronized void writeBackNewMarkingToPetriNet(boolean finished) {
		if (placeMap == null)
			return;
		
		HashSet<Place> changedPlaces = new HashSet<Place>();
		for (Transition t : firedTransitions)
			changedPlaces.addAll(t.getPlaces());
		
		pn.getGraph().getListenerManager().transactionStarted(this);
		try {
			if (!finished) {
				for (Place p : changedPlaces) {
					p.clearTokens();
					for (Object o : placeMap.getPlaceR(p).getTokens()) {
						if (o instanceof TokenDiscrete)
							p.addToken((TokenDiscrete) o);
						else
							if (o instanceof TokenContinuous)
								p.addToken((TokenContinuous) o);
							else
								if (o instanceof TokenObject)
									p.addToken((TokenObject) o);
					}
				}
			} else {
				for (Place p : placeMap.getAllPlacesV()) {
					p.clearTokens();
					for (Object o : placeMap.getPlaceR(p).getTokens()) {
						if (o instanceof TokenDiscrete)
							p.addToken((TokenDiscrete) o);
						else
							if (o instanceof TokenContinuous)
								p.addToken((TokenContinuous) o);
							else
								if (o instanceof TokenObject)
									p.addToken((TokenObject) o);
					}
				}
			}
		} finally {
			pn.getGraph().getListenerManager().transactionFinished(this, false);
		}
	}
	
	/**
	 * Write marking to file.
	 *
	 * @param finished, true writes the cached markings and closes the stream
	 */
	private synchronized void writeMarking2File(boolean finished) {
		if (step % 100 == 1) {
			if (step != 1)
				writer.write(sb);
			
			sb = new StringBuffer();
			writer = new Writer(new File(pathExeAddons).getAbsolutePath() + File.separator + step + ".txt");
			sb.append("step\t");
			for (Place place : placeMap.getAllPlacesV())
				sb.append(place.getName() + "\t");
			
			sb.append("\n");
		}
		sb.append(step + "\t");
		for (Place place : placeMap.getAllPlacesV()) {
			double value = 0;
			for (Object o : placeMap.getPlaceR(place).getTokens()) {
				if (o instanceof TokenDiscrete)
					value += ((TokenDiscrete) o).getValue();
				if (o instanceof TokenContinuous)
					value += ((TokenContinuous) o).getValue();
			}
			sb.append(value + "\t");
		}
		
		sb.append("\n");
		if (finished)
			writer.write(sb);
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#getMarking()
	 */
	public synchronized double[] getMarking() {
		if (placeMap == null)
			return null;
		
		double marking[] = new double[placeMap.getAllPlacesV().size()];
		
		for (Place place : placeMap.getAllPlacesV()) {
			Object[] objs = placeMap.getPlaceR(place).getTokens().toArray();
			int i = pn.getPlacePosition(place);
			for (int j = 0; j < objs.length; j++) {
				if (objs[j] instanceof TokenDiscrete)
					marking[i] += ((TokenDiscrete) objs[j]).getValue();
				if (objs[j] instanceof TokenContinuous)
					marking[i] += ((TokenContinuous) objs[j]).getValue();
			}
		}
		return marking;
	}
	
	/**
	 * Gets the number of discrete tokens.
	 *
	 * @param multiset the multiset
	 * @return the number of discrete tokens
	 */
	public static int getCountIntegerToken(Multiset multiset) {
		int count = 0;
		for (Object token : multiset) {
			if (token instanceof TokenDiscrete)
				count++;
		}
		
		return count;
	}
	
	/**
	 * Adds the arc-weight and capacity to the rakiura Petri net.
	 *
	 * @param pn the pn
	 */
	private void addArcWeightAndCapacity(final PetriNet pn) {
		// Input-arc: guard and expression
		for (final AbstractArc arc : pn.getAllArcs()) {
			org.rakiura.cpn.AbstractArc arcR = arcMap.getarcR(arc);
			if (arcR instanceof InputArc) {
				final InputArc inputarc = (InputArc) arcR;
				inputarc.setGuard(inputarc.new Guard() {
					@Override
					public boolean evaluate() {
						if (getMultiset().getAny(TokenDiscrete.class) instanceof TokenDiscrete) {
							int countToken = getCountIntegerToken(getMultiset());
							if (countToken >= (int) Double.parseDouble(arc.getArcWeight().getValue().toString())) {
								return true;
							}
						}
						if (getMultiset().getAny(TokenContinuous.class) instanceof TokenContinuous) {
							double valueToken = ((TokenContinuous) getMultiset().getAny(TokenContinuous.class)).getValue();
							double weight = Double.parseDouble(arc.getArcWeight().getValue().toString());
							if (valueToken >= weight) {
								return true;
							}
						}
						return false;
					}
				});
				
				inputarc.setExpression(inputarc.new Expression() {
					@Override
					public void evaluate() {
						if(excludedTransitions.contains(arc.getTransition()))
							return;		//new
						if (getMultiset().getAny(TokenDiscrete.class) instanceof TokenDiscrete) {
							var((int) Double.parseDouble(arc.getArcWeight().getValue().toString()));
						}
						if (getMultiset().getAny(TokenContinuous.class) instanceof TokenContinuous) {
							var(inputarc.getID());
						}
					}
				});
			}
		}
		
		// Transitions: guard and action
		for (final Transition transition : pn.getAllTransitions()) {
			org.rakiura.cpn.Transition t = transitionMap.getTransitionR(transition);
			
			if (pn.getAllPlacesWithCapacity().size() > 0) {
				t.setGuard(t.new Guard() {
					@Override
					public boolean evaluate() {						
						return pn.getEnabledTransitionsFromMarking(getMarking(), PetriNet.TRANSITION_RULE)[pn.getTransitionPosition(transition)];
					}
				});
			}
			
			// bordered transition
			if (transition.getInArcs().size() == 0) {
				t.setGuard(t.new Guard() {
					@Override
					public boolean evaluate() {
						int pos = pn.getTransitionPosition(transition);
						return pn.getEnabledTransitionsFromMarking(getMarking(), PetriNet.TRANSITION_RULE)[pos];
					}
				});
			}
			
			t.setAction(t.new Action() {
				@Override
				public void execute() {					
					List<InputArc> inArcs = this.getThisTransition().inputArcs();
					
					// if the transition is a source transition there is no varPool available
					if (getCpnContext().getVarPool().size() == 0) {
						this.getCpnContext().setMultiset(new Multiset());
					}
					
					// side condition: check target place -> has the marking changed, is this transition still enabled; if any of the transitions is no longer enabled -> restore all involved tokens
					if (!isTransitionStillEnabled(getThisTransition())) {
						for (InputArc inArc : inArcs) {
							if ((getMultiset().getAny(TokenContinuous.class) != null) && (getMultiset().getAny(TokenContinuous.class) instanceof TokenContinuous)) {
								double value = ((TokenContinuous) inArc.getPlace().getTokens().getAny()).getValue();
								inArc.getPlace().clearTokens();
								inArc.getPlace().addToken(
										new TokenContinuous(TokenAttributeContinuous.round(value
												+ Double.valueOf(arcMap.getArcV(inArc).getArcWeight().getValue().toString()))));
							} else
								if ((getMultiset().getAny(TokenDiscrete.class) != null) && (getMultiset().getAny(TokenDiscrete.class) instanceof TokenDiscrete)) {
									int count = (int) Double.parseDouble(arcMap.getArcV(inArc).getArcWeight().getValue().toString());
									for (int i = 0; i < count; i++) {
										inArc.getPlace().addToken(new TokenDiscrete(1));
									}
								}
						}
						getThisTransition().setActionText("abort");				
						return;
					}
					
					if ((getMultiset().getAny(TokenContinuous.class) != null) && (getMultiset().getAny(TokenContinuous.class) instanceof TokenContinuous)) {
						
						// create a new token with the value of the arc weight and put the difference back as new token in the input place
						for (InputArc inputArc : inArcs) {
							double value = ((TokenContinuous) this.get(inputArc.getID())).getValue();
							double weight = Double.parseDouble(arcMap.getArcV(inputArc).getArcWeight().getValue().toString());
							double d = TokenAttributeContinuous.round(value - weight);
							
							if (d > 0) {
								// create a new token with the value of the arc weight
								inputArc.getContext().getMultiset().clear();
								inputArc.getContext().getMultiset().add(new TokenContinuous(weight));
								
								inputArc.getPlace().clearTokens();
								inputArc.getPlace().addToken(new TokenContinuous(d));
							}
						}
					}					
				}
			});
		}
		
		// Output-arc: expression
		for (final AbstractArc arc : pn.getAllArcs()) {
			org.rakiura.cpn.AbstractArc arcR = arcMap.getarcR(arc);
			if (arcR instanceof OutputArc) {
				OutputArc outputarc = (OutputArc) arcR;
				outputarc.setExpression(outputarc.new Expression() {
					@Override
					public Multiset evaluate() {		
//						if(excludedTransitions.contains(arc.getTransition()))return new Multiset();		//new
						
						if(getThisArc().getTransition().getActionText().equals("abort"))
							return new Multiset();
						
						// sometimes jfern looses tokens and the multiset is empty, thats why the token are produces with the value of the outgoing arc-weight
						Multiset multiset = new Multiset();
						
						if (getMultiset().getAny(TokenDiscrete.class) instanceof TokenDiscrete) {
							int weight = (int) Double.parseDouble(arc.getArcWeight().getValue().toString());
							for (int i = 0; i < weight; i++) {
								multiset.add(new TokenDiscrete(1));
							}
						}
						else
							if (getMultiset().getAny(TokenContinuous.class) instanceof TokenContinuous) {
								double newTokenValue = Double.parseDouble(arc.getArcWeight().getValue().toString());
								double existingTokenValue = 0;
								
								if (this.getThisArc().getPlace().getTokens().getAny(TokenContinuous.class) != null)
									existingTokenValue = ((TokenContinuous) this.getThisArc().getPlace().getTokens().getAny(TokenContinuous.class)).getValue();
																					
								this.getThisArc().getPlace().clearTokens();// attention: TokenDiscrete are not observed, no problem at the moment 
								
								multiset.add(new TokenContinuous(TokenAttributeContinuous.round(newTokenValue + existingTokenValue)));
							}
						
						// source bordered transition
						if (this.getThisArc().getTransition().inputArcs().size() == 0) {
							PetriNetType type = PetriNetType.getTokenType(placeMap.getAllPlacesV());
							if (type.equals(PetriNetType.CONTINUOUS)) {
								double existingTokenValue = 0;
								double newTokenValue = Double.parseDouble(arc.getArcWeight().getValue().toString());
								
								if (this.getThisArc().getPlace().getTokens().getAny(TokenContinuous.class) != null)
									existingTokenValue = ((TokenContinuous) this.getThisArc().getPlace().getTokens().getAny(TokenContinuous.class)).getValue();
								
								this.getThisArc().getPlace().clearTokens();// attention: TokenDiscrete are not observed, no problem at the moment
								multiset.add(new TokenContinuous(TokenAttributeContinuous.round(newTokenValue + existingTokenValue)));
							}
							else
								if (type.equals(PetriNetType.DISCRETE) || type.equals(PetriNetType.EMPTY)) {
									int arcWeight = (int) Double.parseDouble(arc.getArcWeight().getValue().toString());
									
									for (int i = 0; i < arcWeight; i++) {
										multiset.add(new TokenDiscrete(1));
									}
								}
						}
						return multiset;
					}
				});
			}
		}
	}
	
	/**
	 * Checks if a transition is still enabled.
	 *
	 * @param t the transition (rakiura)
	 * @return true, transition is still enabled
	 */
	private boolean isTransitionStillEnabled(org.rakiura.cpn.Transition t) {
		List<OutputArc> outArcs = t.outputArcs();
		boolean allEnabled = true;
		for (OutputArc outArc : outArcs) {
			double valueTarget = 0;
			for (Object o : outArc.getPlace().getTokens()) {
				if (o instanceof TokenDiscrete)
					valueTarget += ((TokenDiscrete) o).getValue();
				else
					if (o instanceof TokenContinuous)
						valueTarget += ((TokenContinuous) o).getValue();
			}
			
			if (valueTarget + Double.valueOf(arcMap.getArcV(outArc).getArcWeight().getValue().toString()) > Double.valueOf(placeMap.getPlaceV(outArc.getPlace())
					.getCapacity().getValue().toString())) {
				allEnabled = false;
			}
		}
		return allEnabled;
	}
	
	/* (non-Javadoc)
	 * @see org.rakiura.cpn.event.TransitionListener#notify(org.rakiura.cpn.event.TransitionStateChangedEvent)
	 */
	@Override
	public void notify(TransitionStateChangedEvent arg0) {
	}
	
	/* (non-Javadoc)
	 * @see org.rakiura.cpn.event.TransitionListener#notify(org.rakiura.cpn.event.TransitionFinishedEvent)
	 */
	@Override
	public void notify(TransitionFinishedEvent arg0) {
		if (!arg0.getTransition().getActionText().equals("abort")){
			firedTransitions.add(transitionMap.getTransitionV(arg0.getTransition()));	
			
			//write back the new marking for the transition
			Transition transition = transitionMap.getTransitionV(arg0.getTransition());
			Transition transitionReversible = transitionMap.getTransitionV(arg0.getTransition()).getReversibleTransition();
			
			HashSet<Place> changedPlaces = new HashSet<Place>();
			changedPlaces.addAll(transition.getPlaces());
			if(transitionReversible!=null)
				changedPlaces.addAll(transitionReversible.getPlaces());

			for (Place p : changedPlaces) {
				p.clearTokens();
				for (Object o : placeMap.getPlaceR(p).getTokens()) {
					if (o instanceof TokenDiscrete)						
						p.addToken((TokenDiscrete) o);					
					else
						if (o instanceof TokenContinuous)
							p.addToken((TokenContinuous) o);
						else
							if (o instanceof TokenObject)
								p.addToken((TokenObject) o);
				}
			}

			//update exclude list for the reversible transition
			if(transition.getReversibleTransition()!=null){
				Transition random = pn.createRandomReversibleTransition(transitionReversible);								
								
				if(!excludedTransitions.contains(random.getReversibleTransition())){
					excludedTransitions.remove(random);
					excludedTransitions.add(random.getReversibleTransition());
				}
			}
		}
		else
			arg0.getTransition().setActionText("");
	}
	
	/** The fired transitions. */
	private final List<Transition> firedTransitions = new ArrayList<Transition>();
	
	/* (non-Javadoc)
	 * @see org.rakiura.cpn.event.TransitionListener#notify(org.rakiura.cpn.event.TransitionStartedEvent)
	 */
	@Override
	public void notify(TransitionStartedEvent arg0) {		
	}
	
	/* (non-Javadoc)
	 * @see org.rakiura.cpn.event.TransitionListener#notify(org.rakiura.cpn.event.TransitionEvent)
	 */
	@Override
	public void notify(TransitionEvent arg0) {
	}
	
	/** The status. */
	private BackgroundTaskStatusProviderSupportingExternalCallImpl status = new BackgroundTaskStatusProviderSupportingExternalCallImpl("", "");
	
	/** The stringbuffer. */
	private StringBuffer sb;
	
	/** The path to the addon folder. */
	private String pathExeAddons;
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.PetriNetListener#run()
	 */
	@Override
	public void run() {
		status = new BackgroundTaskStatusProviderSupportingExternalCallImpl("", "");
		BackgroundTaskHelper.issueSimpleTaskInWindow("Simulation", "Please wait...", new Runnable() {
			@Override
			public void run() {
				status.setCurrentStatusText1("Build up simulator");
				if (sim == null)
					sim = new Simulator();
				show = false;
				
				sim.addListener(PNJFern.this);
				sim.createSimulator(net, pn);
				sim.setSteps(0);
				
				status.setCurrentStatusText1("Running");
				sim.runSimulation(status);
			}
		}, null, status, false, false);
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.PetriNetListener#run(int)
	 */
	@Override
	public void run(final int steps) {
		status = new BackgroundTaskStatusProviderSupportingExternalCallImpl("", "");
		
		BackgroundTaskHelper.issueSimpleTask("Simulation", "Please wait...", new Runnable() {
			@Override
			public void run() {
				status.setCurrentStatusText1("Build up simulator");
				if (sim == null)
					sim = new Simulator();
				
				sim.addListener(PNJFern.this);
				sim.createSimulator(net, pn);
				sim.setSteps(steps);
				
				status.setCurrentStatusText1("Running");
				sim.runSimulation(status);
			}
		}, null, status);
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.PetriNetListener#stop()
	 */
	@Override
	public void stop() {
		if (sim == null)
			return;
		sim.stop();
		
		if (sb != null)
			writer.write(sb);
	}
	
	/* (non-Javadoc)
	 * @see simulation.SimulatorListener#step(boolean)
	 */
	@Override
	public synchronized void step(boolean finished) {
		if (show && highlightEnabledTransitionsDelay > 0) {
//			pn.getGraph().getListenerManager().transactionStarted(this);
			List<Color> colors = new ArrayList<Color>();
//			try {
				for (Transition transition : firedTransitions) {
					colors.add(AttributeHelper.getFillColor(transition.getNode()));
				}
				
				for (Transition transition : firedTransitions) {
					AttributeHelper.setFillColor(transition.getNode(), Color.RED);
				}
//			} finally {
//				pn.getGraph().getListenerManager().transactionFinished(this, false);
//			}
			try {
				Thread.sleep(highlightEnabledTransitionsDelay);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
//			pn.getGraph().getListenerManager().transactionStarted(this);
//			try {
				int i = 0;
				for (Transition transition : firedTransitions) {
					if(colors.size()>i)
						AttributeHelper.setFillColor(transition.getNode(), colors.get(i++));
				}
//			} finally {
//				pn.getGraph().getListenerManager().transactionFinished(this, false);
//			}
			colors.clear();
		}
		
		if (show || finished) {
//		if (finished) {
//			writeBackNewMarkingToPetriNet(finished);
			GraphHelper.issueCompleteRedrawForGraph(pn.getGraph());
		}
		
		if (write2file)
			writeMarking2File(finished);
		
				
		firedTransitions.clear();
		step++;
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.TransitionListener#fire(vanted.petrinetelements.Transition)
	 */
	@Override
	public void fire(Transition transition) {
		org.rakiura.cpn.Transition t = transitionMap.getTransitionR(transition);
		
		if (t != null && enabled(transition)) {
			t.isEnabled();
			t.fire();
//			writeBackNewMarkingToPetriNet(true);
		}
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#fireTransitions(java.util.List)
	 */
	@Override
	// fires the given transitions (when enabled)
	public void fireTransitions(List<Transition> transitions) {
		boolean[] enabledTr = pn.getEnabledTransitionsFromMarking(getMarking(), PetriNet.TRANSITION_RULE);
		
		List<Transition> copyExcludedTransitionsList = new ArrayList<Transition>(excludedTransitions);
				
		for (Transition t : transitions) {
			if(copyExcludedTransitionsList.contains(t))
				continue;
			
			int pos = pn.getTransitionPosition(t);
			if (enabledTr[pos]) 
				if (transitionMap.getTransitionR(t).isEnabled()) 
					transitionMap.getTransitionR(t).fire();			
		}
		
//		writeBackNewMarkingToPetriNet(true);
		GraphHelper.issueCompleteRedrawForGraph(pn.getGraph());
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.TransitionListener#isEnabled(vanted.petrinetelements.Transition)
	 */
	@Override
	public boolean isEnabled(Transition transition) {
		return enabled(transition);
	}
	
	/**
	 * Checks if the transition is enabled.
	 *
	 * @param transition the transition (vanted)
	 * @return true, transition is enabled
	 */
	private boolean enabled(Transition transition) {
		int pos = pn.getTransitionPosition(transition);
		return pn.getEnabledTransitionsFromMarking(getMarking(), PetriNet.TRANSITION_RULE)[pos];
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#parse2Lib(vanted.petrinetelements.PetriNet)
	 */
	@Override
	public Library parse2Lib(PetriNet pn) {
		return this;
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#getMarking2()
	 */
	@Override
	public Marking getMarking2() {
		Marking marking = new Marking(pn);
		
		for (Place place : placeMap.getAllPlacesV()) {
			marking.addPlaceToken(place);
		}
		return marking;
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#fireTransition(vanted.petrinetelements.Transition)
	 */
	@Override
	public void fireTransition(Transition transition) {
		fire(transition);
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#getSimulator()
	 */
	@Override
	public Simulator getSimulator() {
		return sim;
	}
	
	/**
	 * Creates the library.
	 *
	 * @param petriNet the petri net
	 * @return the library
	 */
	public Library createLibI(PetriNet petriNet) {
		parsePN2JFERN(petriNet);
		return this;
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#highlightEnabledTransitionsDelay(int)
	 */
	@Override
	public void highlightEnabledTransitionsDelay(int ms) {
		highlightEnabledTransitionsDelay = ms;
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#showTokenFlow(boolean)
	 */
	@Override
	public void showTokenFlow(boolean show) {
		this.show = show;
		if(sim!=null)
			sim.setShow(show);			
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#write2File(boolean)
	 */
	@Override
	public void write2File(boolean write) {
		write2file = write;
	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#setEnabledTransitionColor(java.awt.Color)
	 */
	@Override
	public void setEnabledTransitionColor(Color enabledTransitionColor) {	}
	
	/* (non-Javadoc)
	 * @see vanted.petrinetelements.Library#colorEnabledTransitions(boolean)
	 */
	@Override
	public void colorEnabledTransitions(boolean color) {}
}
