/*************************************************************************************
* The Petri-Net Add-on is (c) 2012 Plant Bioinformatics Group, IPK Gatersleben,
* http://bioinformatics.ipk-gatersleben.de
* The source code for this project, is available under the GPL license v2.0 
* available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html. By using this
* Add-on and VANTED you need to accept the terms and conditions of this
* license. For further details see GPL_license.txt in the root folder of this project.
************************************************************************************/

package petri_jfern;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.AttributeHelper;
import org.graffiti.editor.MainFrame;
import org.graffiti.graph.Edge;
import org.graffiti.graph.Graph;
import org.graffiti.graph.Node;

import vanted.Parser;
import vanted.petrinetelements.AbstractArc;
import vanted.petrinetelements.ArcWeight;
import vanted.petrinetelements.NodePN;
import vanted.petrinetelements.NodeShape;

// TODO: Auto-generated Javadoc
/**
 * The Class Translation translates a graph into a Petri net.
 *
 * @author Kevin Pucknat
 */
public class Translation {
	
	/** The graph. */
	private final Graph graph;
	
	/** The nodes. */
	private List<Node> nodes; //not ignored nodes 

	/**
	 * Instantiates a new translation.
	 *
	 * @param graph the graph
	 */
	public Translation(Graph graph) {		
		this.graph = graph;		
	}
	
	/**
	 * Starts the translation.
	 */
	public synchronized void start(){		
		if(graph.getNumberOfDirectedEdges()!=graph.getNumberOfEdges()){
			MainFrame.showMessageDialog("A Petri-Net may not contain undirected arcs/edges!", "Translation Error");
			return;
		}
		
		nodes = new ArrayList<Node>();
		for(Node node : graph.getNodes()){
			NodePN nodePN = new NodePN(node);
			if(!nodePN.getIgnoreNode().getValue())
				nodes.add(node);			
		}
		
		checkShapes(nodes);
	}
	
	/**
	 * Check shapes.
	 *
	 * @param nodes the nodes
	 */
	private void checkShapes(Collection<Node> nodes){		
		List<String> list = new ArrayList<String>();				
		for(Node node : nodes){
			String shape = AttributeHelper.getShape(node);
			if(shape !=null && !shape.equals("")){			
				AttributeHelper.getShapeClassFromShapeName(shape);
				if(!list.contains(shape))
					list.add(shape);
			}			
		}

		autoTranslate();			
	}
	
	/** The edges. */
	Collection<Edge> edges = new ArrayList<Edge>();
	
	/**
	 * Translates the graph, all nodes are changed in places, between that new places a transition is created. The place-transition-place composite is connected with two directed arcs. The two arcs have the same direction as the source edge. 
	 */
	private void autoTranslate(){		
		for(Node node : nodes){
			AttributeHelper.setShape(node, NodeShape.Place.getTxt());			
		}
		
		//without copy of the edge list -> ConcurrentModificationException
		edges.clear();
		for(Edge edge : graph.getEdges()){
			edges.add(edge);
			AttributeHelper.removeEdgeBends(edge);
		}
		
		
		for(Edge edge : edges){			
			Node source = edge.getSource();			
			Node target = edge.getTarget();			
									
			if(!nodes.contains(source))
				continue;
			if(!nodes.contains(target))
				continue;
			
			Node transition = graph.addNode(AttributeHelper.getDefaultGraphicsAttributeForNode(0, 0));
			AttributeHelper.setShape(transition, NodeShape.Transition.getTxt());			
			
			Point2D ps = AttributeHelper.getPosition(source);
			Point2D pt = AttributeHelper.getPosition(target);
			int x = (int) ((ps.getX()+pt.getX())/2);
			int y = (int) ((ps.getY()+pt.getY())/2);
			AttributeHelper.setPosition(transition, x, y);			
						
			graph.addEdgeCopy(edge, source, transition);
			graph.addEdgeCopy(edge, transition, target);	
			graph.deleteEdge(edge);
		}
	
	}
	
	/**
	 * Generates arc attributes from edge labels.
	 */
	public void labels2PetriNetAttributes(){
		for(Edge edge :graph.getEdges()){
			if(AttributeHelper.getLabel(edge, "").isEmpty())continue;

			Object o = Parser.parse(AttributeHelper.getLabel(edge, ""));
			
			if(o!=null){
				if(o instanceof Integer)
					new AbstractArc(edge){}.addArcWeight(new ArcWeight(Integer.parseInt(o.toString()))) ;
				if(o instanceof Double)
					new AbstractArc(edge){}.addArcWeight(new ArcWeight(Double.parseDouble(o.toString()))) ;
			}
								
		}
	}
}
