/*************************************************************************************
* The Petri-Net Add-on is (c) 2012 Plant Bioinformatics Group, IPK Gatersleben,
* http://bioinformatics.ipk-gatersleben.de
* The source code for this project, is available under the GPL license v2.0 
* available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html. By using this
* Add-on and VANTED you need to accept the terms and conditions of this
* license. For further details see GPL_license.txt in the root folder of this project.
************************************************************************************/

package simulation;

import java.util.ArrayList;
import java.util.List;

import org.rakiura.cpn.BasicSimulator;
import org.rakiura.cpn.Net;

import vanted.petrinetelements.PetriNet;

import de.ipk_gatersleben.ag_nw.graffiti.GraphHelper;
import de.ipk_gatersleben.ag_nw.graffiti.services.task.BackgroundTaskStatusProviderSupportingExternalCallImpl;

// TODO: Auto-generated Javadoc
/**
 * The Class Simulator controls the simulation.
 *
 * @author Kevin Pucknat
 */
public class Simulator {
	
	/** The listener. */
	private final ArrayList<SimulatorListener> listener;
	
	/** The simulator. */
	private BasicSimulator simulator;
	
	/** The stop. */
	private boolean stop;
	
	/** The steps. */
	private int steps = 0;
	
	/** The sleep time. */
	private int sleepTime = 1000;

	private PetriNet pn;

	private boolean show = true;

	private boolean isShowSet;

	private boolean infinityRun;
	
	/**
	 * Instantiates a new simulator.
	 */
	public Simulator() {
		listener = new ArrayList<SimulatorListener>();
	}
	
	/**
	 * Creates the simulator.
	 *
	 * @param net the net
	 */
	public void createSimulator(Net net, PetriNet pn) {
		this.pn = pn;
		simulator = new BasicSimulator(net);
	}
	
	/**
	 * Set show.
	 *
	 * @param show the show
	 */
	public void setShow(boolean show) {
		this.show = show;		
	}
	
	/**
	 * Stops the simulation and calls all registered listener (step(true) = last step).
	 */
	public void stop() {
		stop = true;
		for (SimulatorListener l : listener)
			l.step(true);
	}
	
	/**
	 * Runs the simulation for the defined steps and paused the simulation for the defined sleep time. If no more transitions are enabled the simulations stops automatically.
	 *
	 * @param status the status
	 */
	public void runSimulation(BackgroundTaskStatusProviderSupportingExternalCallImpl status) {
		stop = false;
		if (simulator == null)
			return;
		
		for (int i = 1; (i <= steps || steps == 0) && !stop; i++) {
			status.setCurrentStatusText2("Step: " + i);			
			if (!infinityRun && show) //not a infinity run and show token flow
				try {
					Thread.sleep(sleepTime);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			if(!show && !isShowSet){
				pn.getGraph().getListenerManager().transactionStarted(this);			
				isShowSet = true;
			}

			if (!simulator.step() || i == steps || status.wantsToStop()) {	//stop
				try{	
				}finally{
					isShowSet=false;
					pn.getGraph().getListenerManager().transactionFinished(this, false);
				}			
				stop();
				simulator.stop();
				status.setCurrentStatusText1("<html>Simulation finished after " + i + " steps,<br> press OK to close the window.");
				status.setCurrentStatusText2("");
			} else{	//run				
				if (!(steps == 0)) //not a infinity run
					for (SimulatorListener l : listener)
						l.step(false);
			}			
		}
	}
	
	/**
	 * Sets the steps. 0 = infinity steps 
	 *
	 * @param steps the new steps
	 */
	public void setSteps(int steps) {
		this.steps = steps;
		if(steps==0)
			infinityRun=true;
		else
			infinityRun=false;				
	}
	
	/**
	 * Adds the listener.
	 *
	 * @param listener the listener
	 */
	public void addListener(SimulatorListener listener) {
		if (!this.listener.contains(listener))
			this.listener.add(listener);
	}
	
	/**
	 * Removes the listener.
	 *
	 * @param listener the listener
	 */
	public void removeListener(SimulatorListener listener) {
		this.listener.remove(listener);
	}
	
	/**
	 * Gets the listener.
	 *
	 * @return the listener
	 */
	public List<SimulatorListener> getListener() {
		return listener;
	}
	
	/**
	 * Gets the sleep time.
	 *
	 * @return the sleep time
	 */
	public int getSleepTime() {
		return sleepTime;
	}
	
	/**
	 * Sets the sleep time.
	 *
	 * @param sleepTime the new sleep time
	 */
	public void setSleepTime(int sleepTime) {
		if (sleepTime > 0 && sleepTime < 10000)
			this.sleepTime = sleepTime;
	}
	
}
