/*************************************************************************************
* The Petri-Net Add-on is (c) 2012 Plant Bioinformatics Group, IPK Gatersleben,
* http://bioinformatics.ipk-gatersleben.de
* The source code for this project, is available under the GPL license v2.0 
* available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html. By using this
* Add-on and VANTED you need to accept the terms and conditions of this
* license. For further details see GPL_license.txt in the root folder of this project.
************************************************************************************/

package vanted.petrinetelements;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;

import org.AttributeHelper;
import org.color.ColorUtil;
import org.graffiti.attributes.Attribute;
import org.graffiti.graph.Edge;

import vanted.attribute.ArcWeightAttribute;
import vanted.attribute.DrawingEdgeAttribute;
import vanted.attribute.cpn.expression.ExpressionAttribute;
import vanted.attribute.token.TokenAttributeDiscrete;
import vanted.petrinetelements.cpn.Expression;
import vanted.petrinetelements.misc.DrawingEdge;

// TODO: Auto-generated Javadoc
/**
 * The Class AbstractArc is the abstract class for arc handling.
 *
 * @author Kevin Pucknat
 */
public abstract class AbstractArc {
	
	/** The edges. */
	protected List<Edge> edges = new ArrayList<Edge>();
	
	/** The place. */
	protected Place place;
	
	/** The transition. */
	protected Transition transition;
	
	/**
	 * Instantiates a new abstract arc.
	 *
	 * @param edge the edge
	 */
	public AbstractArc(Edge edge) {
		edges.add(edge);
	}
	
	/**
	 * Instantiates a new abstract arc.
	 *
	 * @param place the place
	 * @param transition the transition
	 * @param edge the edge
	 */
	public AbstractArc(Place place, Transition transition, Edge edge) {
		this.place = place;
		this.transition = transition;
		edges.add(edge);
	}
	
	/**
	 * Instantiates a new abstract arc.
	 *
	 * @param transition the transition
	 * @param place the place
	 * @param edge the edge
	 */
	public AbstractArc(Transition transition, Place place, Edge edge) {
		this.place = place;
		this.transition = transition;
		edges.add(edge);
	}
	
	/**
	 * Gets the iD.
	 *
	 * @return the iD
	 */
	public long getID() {
		return edges.get(0).getID();
	}
	
	/**
	 * Gets the place.
	 *
	 * @return the place
	 */
	public Place getPlace() {
		return place;
	}
	
	/**
	 * Gets the transition.
	 *
	 * @return the transition
	 */
	public Transition getTransition() {
		return transition;
	}
	
	/**
	 * Sets the transition.
	 *
	 * @param transition the new transition
	 */
	public void setTransition(Transition transition) {
		this.transition = transition;
	}
	
	/**
	 * Sets the place.
	 *
	 * @param place the new place
	 */
	public void setPlace(Place place) {
		this.place = place;
	}
	
	/**
	 * Gets the edge.
	 *
	 * @return the edge
	 */
	public Edge getEdge() {
		return edges.get(0);
	}
	
	/**
	 * Gets the edges.
	 *
	 * @return the edges
	 */
	public List<Edge> getEdges() {
		return new ArrayList<Edge>(edges);
	}
	
	/**
	 * Sets the edge.
	 *
	 * @param edge the new edge
	 */
	public void setEdge(Edge edge) {
		edges.set(0, edge);
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return edges.get(0).hashCode();
	}
	
	// //expression attribute
	/**
	 * Gets the expression attribute.
	 *
	 * @return the expression attribute
	 */
	private ExpressionAttribute getExpressionAttribute() {
		ExpressionAttribute ea;
		if (!AttributeHelper.hasAttribute(edges.get(0),
				ExpressionAttribute.path, ExpressionAttribute.name)) {
			ea = (ExpressionAttribute) AttributeHelper.getAttributeValue(edges
					.get(0), ExpressionAttribute.path,
					ExpressionAttribute.name, new ExpressionAttribute(
							ExpressionAttribute.name), new ExpressionAttribute(
							TokenAttributeDiscrete.name), true);
		} else {
			ea = (ExpressionAttribute) AttributeHelper.getAttribute(
					edges.get(0), ExpressionAttribute.path + "."
							+ ExpressionAttribute.name);
		}
		return ea;
	}
	
	/**
	 * Adds the expression.
	 *
	 * @param expression the expression
	 */
	public void addExpression(Expression expression) {
		ExpressionAttribute ea = getExpressionAttribute();
		ea.addExpression(expression);
	}
	
	/**
	 * Removes the expression.
	 *
	 * @param expression the expression
	 */
	public void removeExpression(Expression expression) {
		ExpressionAttribute ea = getExpressionAttribute();
		ea.removeExpression(expression);
	}
	
	/**
	 * Removes all the expressions.
	 */
	public void removeAllExpressions() {
		ExpressionAttribute ea = getExpressionAttribute();
		ea.removeAllExpressions();
	}
	
	/**
	 * Gets the expressions.
	 *
	 * @return the expressions
	 */
	public List<Expression> getExpressions() {
		ExpressionAttribute ea = getExpressionAttribute();
		return ea.getExpressions();
	}
	
	// //Arc-weight attribute
	/**
	 * Gets the ArcWeightAttribute which contains the object ArcWeight.
	 *
	 * @return the arc weight attribute
	 */
	private ArcWeightAttribute getArcWeightAttribute() {
		if (!AttributeHelper.hasAttribute(edges.get(0),
				ArcWeightAttribute.path, ArcWeightAttribute.name))
			return null;
		else
			return (ArcWeightAttribute) AttributeHelper.getAttribute(
					edges.get(0), ArcWeightAttribute.path + "."
							+ ArcWeightAttribute.name);
	}
	
	/**
	 * Creates the ArcWeightAttribute which contains the object ArcWeight.
	 *
	 * @return the arc weight attribute
	 */
	private ArcWeightAttribute createArcWeightAttribute() {
		ArcWeightAttribute a = null;
		int i = 0;
		for (Edge edge : edges) {
			AttributeHelper.getAttributeValue(edge, ArcWeightAttribute.path,
					ArcWeightAttribute.sizeAttributeName, 18, 18, true);
			AttributeHelper.getAttributeValue(edge, ArcWeightAttribute.path,
					ArcWeightAttribute.colorAttributeName,
					ColorUtil.getHexFromColor(Color.BLUE), "", true);
			if (i == 0)
				a = (ArcWeightAttribute) AttributeHelper.getAttributeValue(
						edge, ArcWeightAttribute.path, ArcWeightAttribute.name,
						new ArcWeightAttribute(ArcWeightAttribute.name),
						new ArcWeightAttribute(ArcWeightAttribute.name), true);
			i++;
		}
		return a;
	}
	
	/**
	 * Adds the object ArcWeight to the ArcWeightAttribute.
	 *
	 * @param arcWeight the arc weight
	 */
	public void addArcWeight(ArcWeight arcWeight) {
		ArcWeightAttribute ea = getArcWeightAttribute();
		if (ea == null)
			ea = createArcWeightAttribute();
		ea.addArcWeight(arcWeight);
		
		if (edges.size() > 1) {
			for (int i = 1; i < edges.size(); i++) {
				AbstractArc temp = new AbstractArc(edges.get(i)) {};
				temp.addArcWeight(arcWeight);
			}
		}
	}
	
	/**
	 * Removes the object ArcWeight which contains an object with the arc-weight and the ArcWeightAttribute.
	 */
	public void removeArcWeight() {
		ArcWeightAttribute ea = getArcWeightAttribute();
		if (ea == null)
			ea = createArcWeightAttribute();
		ea.removeArcWeight();
		ea = null;
	}
	
	/**
	 * Gets the object ArcWeight which contains an object with the arc-weight.
	 *
	 * @return the arc weight
	 */
	public ArcWeight getArcWeight() {
		ArcWeightAttribute aa = getArcWeightAttribute();
		if (aa == null)
			return new ArcWeight(1);
		return aa.getArcWeight();
	}
	
	// //set drawing attribute
	/**
	 * Gets the DrawingEdgeAttribute which contains the object DrawingEdge.
	 *
	 * @return the draw attribute
	 */
	private DrawingEdgeAttribute getDrawAttribute() {
		if (!AttributeHelper.hasAttribute(edges.get(0),
				DrawingEdgeAttribute.path, DrawingEdgeAttribute.name))
			return null;
		else
			return (DrawingEdgeAttribute) AttributeHelper.getAttribute(
					edges.get(0), DrawingEdgeAttribute.path
							+ Attribute.SEPARATOR + DrawingEdgeAttribute.name);
	}
	
	/**
	 * Adds the object DrawingEdge to the DrawingEdgeAttribute.
	 *
	 * @param draw the draw
	 */
	public void addDraw(DrawingEdge draw) {
		DrawingEdgeAttribute da = getDrawAttribute();
		if (da == null)
			da = createDrawAttribute();
		da.addDraw(draw);
	}
	
	/**
	 * Creates the DrawingEdgeAttribute which contains the object DrawingEdge.
	 *
	 * @return the drawing edge attribute
	 */
	private DrawingEdgeAttribute createDrawAttribute() {
		return (DrawingEdgeAttribute) AttributeHelper.getAttributeValue(edges
				.get(0), DrawingEdgeAttribute.path, DrawingEdgeAttribute.name,
				new DrawingEdgeAttribute(DrawingEdgeAttribute.name),
				new DrawingEdgeAttribute(DrawingEdgeAttribute.name), true);
	}
	
	/**
	 * Removes the object DrawingEdge which contains a String with a label and the DrawingEdgeAttribute.
	 */
	public void removeDraw() {
		DrawingEdgeAttribute da = getDrawAttribute();
		if (da != null) {
			da.removeDraw();
			for (Edge edge : edges) {
				AttributeHelper.deleteAttribute(edge,
						DrawingEdgeAttribute.path, DrawingEdgeAttribute.name);
			}
			AttributeHelper.deleteAttribute(edges.get(0),
					DrawingEdgeAttribute.path, DrawingEdgeAttribute.name);
		}
	}
	
	/**
	 * Gets the object DrawingEdge which contains a String with a label.
	 *
	 * @return the draw
	 */
	public DrawingEdge getDraw() {
		DrawingEdgeAttribute da = getDrawAttribute();
		if (da == null)
			return new DrawingEdge("");
		return da.getDraw();
	}
}
