/*************************************************************************************
* The Petri-Net Add-on is (c) 2012 Plant Bioinformatics Group, IPK Gatersleben,
* http://bioinformatics.ipk-gatersleben.de
* The source code for this project, is available under the GPL license v2.0 
* available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html. By using this
* Add-on and VANTED you need to accept the terms and conditions of this
* license. For further details see GPL_license.txt in the root folder of this project.
************************************************************************************/

package vanted.petrinetelements;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Random;

import misc.MatrixOperation;

import org.AttributeHelper;
import org.ErrorMsg;
import org.graffiti.editor.MainFrame;
import org.graffiti.graph.Edge;
import org.graffiti.graph.Graph;
import org.graffiti.graph.Node;

import petri_jfern.PNJFern;
import petri_jfern.Translation;
import pipe.analysis.Matrix;
import vanted.attribute.HideArcWeightAttribute;
import vanted.attribute.HideCapacityAttribute;
import vanted.attribute.HideLabelAttribute;
import vanted.attribute.HideTokenAttribute;
import vanted.petrinetelements.cpn.Marking;
import vanted.petrinetelements.misc.Check;
import vanted.petrinetelements.misc.HelperPNMergeNodes;
import vanted.petrinetelements.misc.HideArcWeight;
import vanted.petrinetelements.misc.HideCapacity;
import vanted.petrinetelements.misc.HideLabel;
import vanted.petrinetelements.misc.HideToken;
import vanted.validation.Validator;

// TODO: Auto-generated Javadoc
/**
 * The Class PetriNet for Petri net handling.
 *
 * @author Kevin Pucknat
 */
public class PetriNet{	
	
	/** The capacity maximum value. */
	public static Integer CAPACITY_MAX_VALUE = Integer.MAX_VALUE;
	
	/** The transition rule. */
	public static TransitionRule TRANSITION_RULE = TransitionRule.STRONG_WITH_SIDE_CONDITION;
	
	/** The graph. */
	private Graph graph;
	
	/** The places. */
	private PlaceList places;
	
	/** The transitions. */
	private List<Transition> transitions;
	
	/** The arcs. */
	private List<AbstractArc> arcs;
	
	/** The listener. */
	private List<PetriNetListener> listener;		
	
	/** The capacity vector. */
	private double[] capacityVector;
	
	/** The mo. */
	private MatrixOperation mo;
	
	/**
	 * Instantiates a new petri net.
	 */
	public PetriNet() {	
		this(null);
	}
	
	/**
	 * Instantiates a new petri net.
	 *
	 * @param graph the graph
	 */
	public PetriNet(Graph graph) {
		this.graph = graph;		
		places = new PlaceList();
		transitions = new ArrayList<Transition>();
		arcs = new ArrayList<AbstractArc>();
		listener = new ArrayList<PetriNetListener>();	
		capacityVector = new double[0]; 
	}

	
	/**
	 * Gets the graph.
	 *
	 * @return the graph
	 */
	public Graph getGraph(){
		return graph;
	}
		
	/**
	 * Creates the Petri net if it is correct.
	 *
	 * @param graph the graph
	 * @param transformPetriNet the transform petri net
	 * @return true, if successful
	 */
	public synchronized boolean createPetriNet(Graph graph, boolean transformPetriNet){
		this.graph = graph;		
		if (graph != null){
			if(new Validator().isPNCorrectSilent(graph)){//if the vanted graph is a valid Petri-Net, create a real Petri-Net				
				if(createPetriNet2(graph, transformPetriNet)){
					capacityVector = createCapacityArray(places);					
					return true;
				}else return false;
			}
			else{
				ErrorMsg.addErrorMessage("No valid graph!");
				return false;
			}
		}else return false;		
	}
	
	/**
	 * Creates the Petri net structure, checks for capacity, connects the logical places with its belonging transitions and splits up the reversible edges.
	 *
	 * @param graph the graph
	 * @param transformPetriNet the transform petri net
	 * @return true, if successful
	 */
	private boolean createPetriNet2(Graph graph, boolean transformPetriNet){
		Translation translate = new Translation(graph);
		translate.labels2PetriNetAttributes();
		
		for(Node node : graph.getNodes()){
			NodePN nodePN = new NodePN(node);
			if(!nodePN.getIgnoreNode().getValue()){			
				if(Transition.isTransition(node)){				
					Transition transition = new Transition(node);
					transitions.add(transition);	
					//in
					for(Edge edge : node.getDirectedInEdges()){
						if(!places.contains(edge.getSource())){							
							Place place = new Place(edge.getSource());
							if(place.getIgnoreNode().getValue())continue;
							places.add(place);
							InArc inArc = new InArc(place, transition, edge);
							arcs.add(inArc);
							
							if(!isTokenCapacityCorrect(place)) return false;
						}else{
							Place place = places.get(places.indexOf(edge.getSource()));
							InArc inArc = new InArc(place, transition, edge);
							arcs.add(inArc);
						}
							
					}
					
					//out
					for(Edge edge : node.getDirectedOutEdges()){
						if(!places.contains(edge.getTarget())){
							Place place = new Place(edge.getTarget());
							if(place.getIgnoreNode().getValue())continue;
							places.add(place);
							OutArc outArc = new OutArc(transition, place, edge);
							arcs.add(outArc);
							
							if(!isTokenCapacityCorrect(place)) return false;
						}else{
							Place place = places.get(places.indexOf(edge.getTarget()));
							OutArc outArc = new OutArc(transition, place, edge);							
							arcs.add(outArc);
						}
					}
				}	
			}
		}

		if(transformPetriNet){
			//merge nodes with same labels
			HelperPNMergeNodes.mergeNodesWithSameLabel(graph.getNodes(), this);		

			//transform reversible edges 
			boolean reversibleIn = false;
			boolean reversibleOut = false;
			List<Transition> tempTransition = new ArrayList<Transition>();
			
			for(Transition transition : transitions){
				reversibleIn = false;
				reversibleOut = false;
				Transition t = new Transition(transition.getNode());		
				List<InArc> tempInArc = new ArrayList<InArc>();
				List<OutArc> tempOutArc = new ArrayList<OutArc>();
				
				for(InArc inArc : transition.getInArcs()){
					if(Check.isReversibleEdge(inArc.getEdge())){
						tempOutArc.add(new OutArc(t, inArc.getPlace(), inArc.getEdge()));					
						reversibleIn=true;
					}
				}		
				
				for(OutArc outArc : transition.getOutArcs()){				
					if(Check.isReversibleEdge(outArc.getEdge())){
						tempInArc.add(new InArc(outArc.getPlace(), t, outArc.getEdge()));					
						reversibleOut=true;
					}
				}
				
				if(reversibleIn && reversibleOut){
					arcs.addAll(tempInArc);
					arcs.addAll(tempOutArc);
					t.setRevTransition(true);
										
					t.setReversibleTransition(transition);
					transition.setReversibleTransition(t);
					
					tempTransition.add(t);	
				}
							
			}			
			transitions.addAll(tempTransition);
		}
		
		return true;
	}
		
	
	
	/**
	 * Checks if the token capacity is correct.
	 *
	 * @param place the place
	 * @return true, if is token capacity correct
	 */
	private boolean isTokenCapacityCorrect(Place place){
		double tokenWeight=0;
		for(AbstractToken token : place.getTokens()){
			if(token instanceof TokenDiscrete)							
				tokenWeight+=((TokenDiscrete)token).getValue();
			else if(token instanceof TokenDiscrete)
				tokenWeight+=((TokenContinuous)token).getValue();
		}
		double capacity = Double.parseDouble(place.getCapacity().getValue().toString());
		if(tokenWeight>capacity && capacity!=0){		
			MainFrame.showMessageDialog("The token weight is to height or the place capycity is to low", "Error");
			return false;						
		}
		return true;
	}
	
	/**
	 * Gets all the places.
	 *
	 * @return the all places
	 */
	public List<Place> getAllPlaces(){		
		return new ArrayList<Place>(places);
	}
	
	/**
	 * Gets all the transitions.
	 *
	 * @return the all transitions
	 */
	public List<Transition> getAllTransitions(){		
		return new ArrayList<Transition>(transitions);
	}
	
	/**
	 * Gets the all arcs.
	 *
	 * @return the all arcs
	 */
	public List<AbstractArc> getAllArcs(){		
		return new ArrayList<AbstractArc>(arcs);
	}
	
	/**
	 * Removes this node.
	 *
	 * @param node the node
	 * @return true, if successful
	 */
	public boolean removeNode(NodePN node){		
		if(places.remove(node))return true;		
		else if(transitions.remove(node))return true;
		
		return false;
	}
	
	/**
	 * Gets the marking.
	 *
	 * @return the marking
	 */
	public Marking getMarking(){
		Marking marking = new Marking(this);
		
		for(Place place : places){
			marking.addPlaceToken(place);
		}
		
		return marking;
	}
	
	/**
	 * Gets all the places with capacity.
	 *
	 * @return the all places with capacity
	 */
	public List<Place> getAllPlacesWithCapacity(){
		List<Place> temp = new ArrayList<Place>();
		for(Place place : places){			
			if(Double.parseDouble(place.getCapacity().getValue().toString())!=0)
				temp.add(place);				
		}
		
		return temp;
	}

	/**
	 * Gets all the selected transitions.
	 *
	 * @param nodes the nodes
	 * @return the all selected transitions
	 */
	public List<Transition> getAllSelectedTransitions(Collection<Node> nodes){
		List<Transition> activeTransitions = new ArrayList<Transition>();
//		int revTransitions=0;
		for(Node node : nodes){			
			for(Transition transition : transitions){				
				if(transition.getNode().equals(node)){					
//					if (transition.isRevTransition()) 
//						revTransitions++;					
					activeTransitions.add(transition);					
				}
			}			
		}
				
//		if(revTransitions>0)
//			return createRandomReversibleTransitions(activeTransitions, revTransitions);
		
		return activeTransitions;
	}
	
		
	/**
	 * Returns one random transition of the pair of reversible transitions. 
	 *
	 * @param transitions the reversible transition
	 * @return only one of a pair of reversible transitions 
	 */
	public Transition createRandomReversibleTransition(Transition transition){
		if(transition.getReversibleTransition()==null)return transition;
		
		boolean isTransitionEnabled = getEnabledTransitionsFromMarking(getMarking(this), TRANSITION_RULE)[getTransitionPosition(transition)];					

		Random rand = new Random();
		if(rand.nextBoolean()){
			if(isTransitionEnabled)
				return transition;
			else
				return transition.getReversibleTransition();
		}else{
			if(getEnabledTransitionsFromMarking(getMarking(this), TRANSITION_RULE)[getTransitionPosition(transition.getReversibleTransition())])								
				return transition.getReversibleTransition();			
			else
				return transition;
		}
	}
	
	/**
	 * Returns one random transition of the pair of reversible transitions. 
	 *
	 * @param transitions the transitions
	 * @return only one of a pair of reversible transitions 
	 */
	public List<Transition> createRandomReversibleTransitions(List<Transition> transitions){
		List<Transition> randomTransitions = new ArrayList<Transition>();

		for(Transition transition : transitions)		
			if(transition.isRevTransition())
				randomTransitions.add(createRandomReversibleTransition(transition));			
				
		return randomTransitions;		
	}
	
	//infinity run
	/**
	 * Run infinity.
	 */
	public void run(){
		for(PetriNetListener l : listener){
			l.run();
		}
	}
	
	//runs ... steps if any transition enabled
	/**
	 * Run with ... steps.
	 *
	 * @param steps the steps
	 */
	public void run(int steps){
		for(PetriNetListener l : listener){
			l.run(steps);
		}
	}
	
	//stops the two run methods
	/**
	 * Stop both run methods.
	 */
	public void stop(){
		for(PetriNetListener l : listener){
			l.stop();
		}
	}
	
	
	/**
	 * Adds the listener.
	 *
	 * @param listener the listener
	 */
	public void addListener(PetriNetListener listener){		
		if(!this.listener.equals(listener))
			this.listener.add(listener);
	}
	
	/**
	 * Gets the listeners.
	 *
	 * @return the listeners
	 */
	public List<PetriNetListener> getListeners(){
		return listener;
	}
			
	/**
	 * Gets the enabled transitions.
	 *
	 * @param rule the rule
	 * @return the enabled transitions
	 */
	public boolean[] getEnabledTransitions(TransitionRule rule){
		double marking[] = new double[places.size()];		
		int i=0;
		for(Place place : places){			
			for(AbstractToken token : place.getTokens()){				
				if(token instanceof TokenDiscrete)
					marking[i]+=((TokenDiscrete)token).getValue();
				if(token instanceof TokenContinuous)
					marking[i]+=((TokenContinuous)token).getValue();				
			}	
			i++;
		}		
		return getEnabledTransitionsFromMarking(marking, rule);
	}
	

	/**
	 * Gets the enabled transitions from marking.
	 *
	 * @param marking the marking
	 * @param rule the rule
	 * @return the enabled transitions from marking
	 */
	public boolean[] getEnabledTransitionsFromMarking(double[] marking, TransitionRule rule){
		switch(rule){
			case WEAK: return getEnabledTransitionsFromMarkingWeak(marking);			
			case STRONG: return getEnabledTransitionsFromMarkingStrong1(marking);			
			case STRONG_WITH_SIDE_CONDITION: return getEnabledTransitionsFromMarkingStrong2(marking);			
		}
		return null;
	}
	
	/**
	 * Gets the enabled transitions from marking with the weak transition rule.
	 *
	 * @param marking the marking
	 * @return the enabled transitions from marking weak
	 */
	private boolean[] getEnabledTransitionsFromMarkingWeak(double[] marking){
		if(mo==null)mo = new MatrixOperation(this);
		
		boolean enabled[] = new boolean[transitions.size()];
		for(int i=0; i<enabled.length;i++){
			enabled[i]=true;
		}					
		
		Matrix incidence = mo.getIncidenceMatrix();
		double bMatrix[][] = mo.getBMatrix();
		
		for(int m=0; m<incidence.getColumnDimension(); m++){ //transitions
			for(int n=0; n<incidence.getRowDimension(); n++){ //places
				if(marking[n]!=-1 && //omega
						!(marking[n]>=bMatrix[n][m])){ //pre-condition
					enabled[m]=false;
				}
			}
		}
		
		return enabled;		
	}
	
	//without side conditions
	/**
	 * Gets the enabled transitions from marking with the strong transition rule.
	 *
	 * @param marking the marking
	 * @return the enabled transitions from marking strong1
	 */
	private boolean[] getEnabledTransitionsFromMarkingStrong1(double[] marking){
		if(mo==null)mo = new MatrixOperation(this);
		
		boolean enabled[] = new boolean[transitions.size()];
		for(int i=0; i<enabled.length;i++){
			enabled[i]=true;
		}					

		Matrix incidence = mo.getIncidenceMatrix();
		double bMatrix[][] = mo.getBMatrix();
		double fMatrix[][] = mo.getFMatrix();
		
		for(int t=0; t<incidence.getColumnDimension(); t++){ //transitions
			for(int p=0; p<incidence.getRowDimension(); p++){ //places			
				if(marking[p]!=-1 &&	//omega
					((marking[p]<bMatrix[p][t]) ||	//pre-condition
						!(marking[p]<=capacityVector[p]-fMatrix[p][t]))){	//post-condition
					enabled[t]=false;
				}
			}
		}
		
		return enabled;			
	}
	
	//with side conditions
	/**
	 * Gets the enabled transitions from marking with the strong transition rule with side conditions.
	 *
	 * @param marking the marking
	 * @return the enabled transitions from marking strong2
	 */
	private boolean[] getEnabledTransitionsFromMarkingStrong2(double[] marking){
		if(mo==null)mo = new MatrixOperation(this);
		
		boolean enabled[] = new boolean[transitions.size()];
		for(int i=0; i<enabled.length;i++){
			enabled[i]=true;
		}					

		Matrix incidence = mo.getIncidenceMatrix();
		double bMatrix[][] = mo.getBMatrix();
		double fMatrix[][] = mo.getFMatrix();
		
		for(int t=0; t<incidence.getColumnDimension(); t++){ //transitions
			for(int p=0; p<incidence.getRowDimension(); p++){ //places			
				if(marking[p]!=-1 &&	//omega
					((marking[p]<bMatrix[p][t]) ||	//pre-condition 
						(marking[p]+(fMatrix[p][t]-bMatrix[p][t]))>capacityVector[p])){	//post-condition						
					enabled[t]=false;
				}
			}
		}
		
		return enabled;			
	}
	
	
	/**
	 * Creates the capacity array.
	 *
	 * @param places the places
	 * @return the double[]
	 */
	private double[] createCapacityArray(List<Place> places){
		double vec[] = new double[places.size()];
	   	int i=0;
	   	for(Place place : places){
	   		vec[i++] = Double.parseDouble(place.getCapacity().getValue().toString());
	   	}	   	
	   	return vec;
	}
	
	/**
	 * Gets the Petri net marking.
	 *
	 * @param pn the pn
	 * @return the marking
	 */
	public static double[] getMarking(PetriNet pn){
		double marking[] = new double[pn.getAllPlaces().size()];
		int i=0;
		for(Place place : pn.getAllPlaces()){
			for(AbstractToken token : place.getTokens()){
				if(token instanceof TokenDiscrete){
					marking[i]+=((TokenDiscrete)token).getValue();
				}
				else if(token instanceof TokenContinuous){
					marking[i]+=((TokenContinuous)token).getValue();
				}
			}
			i++;
		}
		
		return marking;
	}
	
	
	////hide token attribute
	/**
	 * Gets the HideTokenAttribute which contains the object HideToken.
	 *
	 * @return the hide token attribute
	 */
	private HideTokenAttribute getHideTokenAttribute(){	
		if(!AttributeHelper.hasAttribute(graph, HideTokenAttribute.path, HideTokenAttribute.name))
			return null;
		else			
			return (HideTokenAttribute)AttributeHelper.getAttribute(graph, HideTokenAttribute.path+"."+HideTokenAttribute.name);		
	}
		
	/**
	 * Adds the object HideToken to the HideTokenAttribute.
	 *
	 * @param hideToken the hide token
	 */
	public void addHideToken(HideToken hideToken) {		
		HideTokenAttribute ita = getHideTokenAttribute();
		if(ita==null)			
			ita = createHideTokenAttribute();
		ita.addHideToken(hideToken);
	}
	
	/**
	 * Creates the HideTokenAttribute which contains the object HideToken.
	 *
	 * @return the hide token attribute
	 */
	private HideTokenAttribute createHideTokenAttribute() {
		return (HideTokenAttribute) AttributeHelper.getAttributeValue(graph, HideTokenAttribute.path, HideTokenAttribute.name, new HideTokenAttribute(
				HideTokenAttribute.name), new HideTokenAttribute(HideTokenAttribute.name), true);
	}
	
	/**
	 * Removes the object HideToken which contains true (hiding all tokens), or false and the HideTokenAttribute.
	 */
	public void removeHideToken() {		
		HideTokenAttribute a = getHideTokenAttribute();		
		if(a!=null){
			a.removeHideToken();
			AttributeHelper.deleteAttribute(graph, HideTokenAttribute.path, HideTokenAttribute.name);
		}
	}
	
	/**
	 * Gets the object HideToken which contains true (hiding all tokens), or false.
	 *
	 * @return the hide token
	 */
	public HideToken getHideToken(){
		HideTokenAttribute a = getHideTokenAttribute();
		if(a==null)
			return new HideToken(false);
		return a.getHideToken();
	}
	
	////hide capacity attribute
	/**
	 * Gets the HideCapacityAttribute which contains the object HideCapacity.
	 *
	 * @return the hide capacity attribute
	 */
	private HideCapacityAttribute getHideCapacityAttribute(){	
		if(!AttributeHelper.hasAttribute(graph, HideCapacityAttribute.path, HideCapacityAttribute.name))
			return null;
		else			
			return (HideCapacityAttribute)AttributeHelper.getAttribute(graph, HideCapacityAttribute.path+"."+HideCapacityAttribute.name);		
	}
		
	/**
	 * Adds the object HideCapacity to the HideCapacityAttribute.
	 *
	 * @param hideCapacity the hide capacity
	 */
	public void addHideCapacity(HideCapacity hideCapacity) {		
		HideCapacityAttribute a = getHideCapacityAttribute();
		if(a==null)			
			a = createHideCapacityAttribute();
		a.addHideCapacity(hideCapacity);
	}
	
	/**
	 * Creates the HideCapacityAttribute which contains the object HideCapacity.
	 *
	 * @return the hide capacity attribute
	 */
	private HideCapacityAttribute createHideCapacityAttribute() {
		return (HideCapacityAttribute) AttributeHelper.getAttributeValue(graph, HideCapacityAttribute.path, HideCapacityAttribute.name, new HideCapacityAttribute(
				HideCapacityAttribute.name), new HideCapacityAttribute(HideCapacityAttribute.name), true);
	}
	
	/**
	 * Removes the object HideCapacity which contains true (hiding all capacities), or false and the HideCapacityAttribute.
	 */
	public void removeHideCapacity() {		
		HideCapacityAttribute a = getHideCapacityAttribute();		
		if(a!=null){
			a.removeHideCapacity();
			AttributeHelper.deleteAttribute(graph, HideCapacityAttribute.path, HideCapacityAttribute.name);
		}
	}
	
	/**
	 * Get the object HideCapacity which contains true (hiding all capacities), or false.
	 *
	 * @return the hide capacity
	 */
	public HideCapacity getHideCapacity(){
		HideCapacityAttribute a = getHideCapacityAttribute();
		if(a==null)
			return new HideCapacity(false);
		return a.getHideCapacity();
	}
	
	////hide arc weight attribute 
	/**
	 * Gets the HideArcWeightAttribute which contains the object HideArcWeight.
	 *
	 * @return the hide arc weight attribute
	 */
	private HideArcWeightAttribute getHideArcWeightAttribute(){	
		if(!AttributeHelper.hasAttribute(graph, HideArcWeightAttribute.path, HideArcWeightAttribute.name))
			return null;
		else		
			return (HideArcWeightAttribute)AttributeHelper.getAttribute(graph, HideArcWeightAttribute.path+"."+HideArcWeightAttribute.name);		
	}
		
	/**
	 * Adds the object HideArcWeight to the HideArcWeightAttribute.
	 *
	 * @param hideArcWeight the hide arc weight
	 */
	public void addHideArcWeight(HideArcWeight hideArcWeight) {		
		HideArcWeightAttribute a = getHideArcWeightAttribute();
		if(a==null)			
			a = createHideArcWeightAttribute();
		a.addHideArcWeight(hideArcWeight);
	}
	
	/**
	 * Creates the HideArcWeightAttribute which contains the object HideArcWeight.
	 *
	 * @return the hide arc weight attribute
	 */
	private HideArcWeightAttribute createHideArcWeightAttribute() {
		return (HideArcWeightAttribute) AttributeHelper.getAttributeValue(graph, HideArcWeightAttribute.path, HideArcWeightAttribute.name, new HideArcWeightAttribute(
				HideArcWeightAttribute.name), new HideArcWeightAttribute(HideArcWeightAttribute.name), true);
	}
	
	/**
	 * Removes the object HideArcWeight which contains true (hiding all arc weights), or false and the HideArcWeightAttribute.
	 */
	public void removeHideArcWeight() {		
		HideArcWeightAttribute a = getHideArcWeightAttribute();		
		if(a!=null){
			a.removeHideArcWeight();
			AttributeHelper.deleteAttribute(graph, HideArcWeightAttribute.path, HideArcWeightAttribute.name);
		}
	}
	
	/**
	 * Gets the object HideArcWeight which contains true (hiding all arc weight), or false.
	 *
	 * @return the hide arc weight
	 */
	public HideArcWeight getHideArcWeight(){
		HideArcWeightAttribute a = getHideArcWeightAttribute();
		if(a==null)
			return new HideArcWeight(false);
		return a.getHideArcWeight();
	}
	
////hide label attribute 
	/**
 * Gets the HideLabelAttribute which contains the object HideLabel.
 *
 * @return the hide label attribute
 */
private HideLabelAttribute getHideLabelAttribute(){	
		if(!AttributeHelper.hasAttribute(graph, HideLabelAttribute.path, HideLabelAttribute.name))
			return null;
		else			
			return (HideLabelAttribute)AttributeHelper.getAttribute(graph, HideLabelAttribute.path+"."+HideLabelAttribute.name);		
	}
		
	/**
	 * Adds the object HideLabel to the HideLabelAttribute.
	 *
	 * @param hideLabel the hide label
	 */
	public void addHideLabel(HideLabel hideLabel) {		
		if(graph==null)return;
		HideLabelAttribute a = getHideLabelAttribute();
		if(a==null)			
			a = createHideLabelAttribute();
		a.addHideLabel(hideLabel);
	}
	
	/**
	 * Creates the HideLabelAttribute which contains the object HideLabel.
	 *
	 * @return the hide label attribute
	 */
	private HideLabelAttribute createHideLabelAttribute() {
		return (HideLabelAttribute) AttributeHelper.getAttributeValue(graph, HideLabelAttribute.path, HideLabelAttribute.name, new HideLabelAttribute(
				HideLabelAttribute.name), new HideLabelAttribute(HideLabelAttribute.name), true);
	}
	
	/**
	 * Removes the object HideLabel which contains true (hiding all node label), or false and the HideLabelAttribute.
	 */
	public void removeHideLabel() {		
		HideLabelAttribute a = getHideLabelAttribute();		
		if(a!=null){
			a.removeHideLabel();
			AttributeHelper.deleteAttribute(graph, HideLabelAttribute.path, HideLabelAttribute.name);
		}
	}
	
	/**
	 * Gets the object HideLabel which contains true (hiding all node label), or false.
	 *
	 * @return the hide label
	 */
	public HideLabel getHideLabel(){
		HideLabelAttribute a = getHideLabelAttribute();
		if(a==null)
			return new HideLabel(false);
		return a.getHideLabel();
	}
	
	
	/**
	 * Gets the place position in the Petri net.
	 *
	 * @param place the place
	 * @return the place position
	 */
	public int getPlacePosition(Place place){
		for(int i=0; i<places.size(); i++)
			if(places.get(i).equals(place))
				return i;
			
		return -1;
	}
	
	/**
	 * Gets the transition position in the Petri net.
	 *
	 * @param transition the transition
	 * @return the transition position
	 */
	public int getTransitionPosition(Transition transition){
		for(int i=0; i<transitions.size(); i++)
			if(transitions.get(i).equals(transition))
				return i;
			
		return -1;
	}
	
	
	/** The library. */
	private Library library;
	
	/**
	 * Parses the Petri net to the library.
	 *
	 * @return the library
	 */
	public Library parsePetriNet2Lib(){
		PNJFern j = new PNJFern();
		library=j;
		return j.createLibI(this);
	}
	
	/**
	 * Gets the library.
	 *
	 * @return the library
	 */
	public Library getLibrary(){
		return library;
	}
	
	public double[] getCapacityVector(){
		if(capacityVector!=null)
			return capacityVector;
		else
			return createCapacityArray(places);
	}
}
