/*************************************************************************************
* The Petri-Net Add-on is (c) 2012 Plant Bioinformatics Group, IPK Gatersleben,
* http://bioinformatics.ipk-gatersleben.de
* The source code for this project, is available under the GPL license v2.0 
* available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html. By using this
* Add-on and VANTED you need to accept the terms and conditions of this
* license. For further details see GPL_license.txt in the root folder of this project.
************************************************************************************/

package vanted.petrinetelements;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;

import org.AttributeHelper;
import org.graffiti.graph.Node;

import vanted.attribute.cpn.guard.GuardAttribute;
import vanted.attribute.token.TokenAttributeDiscrete;
import vanted.petrinetelements.cpn.Guard;
import vanted.petrinetelements.misc.Identity;

// TODO: Auto-generated Javadoc
/**
 * The Class Transition for transition handling.
 *
 * @author Kevin Pucknat
 */
public class Transition extends NodePN {
	
	/** The in arcs. */
	private List<InArc> inArcs = new ArrayList<InArc>();
	
	/** The out arcs. */
	private List<OutArc> outArcs = new ArrayList<OutArc>();
	
	/** The transition listener. */
	private final List<TransitionListener> transitionListener = new ArrayList<TransitionListener>();
	
	/** Is the transition reversible. */
	private boolean isRevTransition;
	
	/** The reversible transitions. */
	private Transition revTransition;
	
	/**
	 * Instantiates a new transition.
	 *
	 * @param node the node
	 */
	public Transition(Node node) {
		super(node);
		if (getIdentity().getValue() == 0) {
			Random rand = new Random();
			addIdentity(new Identity(rand.nextLong()));
		}
	}
	
	/**
	 * Instantiates a new reversible transition.
	 *
	 * @param node the node
	 */
	public Transition(Transition transition) {
		super(transition.getNode());
		revTransition = transition;
		if (getIdentity().getValue() == 0) {
			Random rand = new Random();
			addIdentity(new Identity(rand.nextLong()));
		}
	}
	
	
	/**
	 * Gets the reversible transition.
	 *	
	 */
	public Transition getReversibleTransition(){
		return revTransition;
	}
	
	/**
	 * Sets a reversible transition.
	 *
	 * @param transition the referenced transition
	 */
	public void setReversibleTransition(Transition transition){
		revTransition = transition;
	}
	
	/**
	 * Adds the out arc of the transition.
	 *
	 * @param arc the arc
	 */
	public void addOutArc(OutArc arc) {
		outArcs.add(arc);
	}
	
	/**
	 * Sets the in arcs of the transition.
	 *
	 * @param inArcs the new in arcs
	 */
	public void setInArcs(List<InArc> inArcs) {
		this.inArcs = inArcs;
	}
	
	/**
	 * Gets the in arcs of the transition.
	 *
	 * @return the in arcs
	 */
	public List<InArc> getInArcs() {
		return inArcs;
	}
	
	/**
	 * Adds the in arc of the transition.
	 *
	 * @param arc the arc
	 */
	public void addInArc(InArc arc) {
		inArcs.add(arc);
	}
	
	/**
	 * Sets the out arcs of the transition.
	 *
	 * @param outArcs the new out arcs
	 */
	public void setOutArcs(List<OutArc> outArcs) {
		this.outArcs = outArcs;
	}
	
	/**
	 * Gets the out arcs of the transition.
	 *
	 * @return the out arcs
	 */
	public List<OutArc> getOutArcs() {
		return outArcs;
	}
	
	/**
	 * Checks if the node is a transition.
	 *
	 * @param nd the nd
	 * @return true, if is transition
	 */
	public static boolean isTransition(Node nd) {
		return NodeShape.isTransition(AttributeHelper.getShape(nd).toLowerCase());
	}
	
	/**
	 * Gets the guard attribute of the transition.
	 *
	 * @param value the value
	 * @param guard the guard
	 * @return the guard attribute
	 */
	private GuardAttribute getGuardAttribute(String value, Guard guard) {
		GuardAttribute ga;
		if (!AttributeHelper.hasAttribute(nodes.get(0), GuardAttribute.path, GuardAttribute.name)) {
			ga = (GuardAttribute) AttributeHelper.getAttributeValue(nodes.get(0), GuardAttribute.path, GuardAttribute.name, new GuardAttribute(
					GuardAttribute.name), new GuardAttribute(TokenAttributeDiscrete.name), true);
		} else {
			ga = (GuardAttribute) AttributeHelper.getAttribute(nodes.get(0), GuardAttribute.path + "." + GuardAttribute.name);
		}
		return ga;
	}
	
	/**
	 * Adds the guard of the transition.
	 *
	 * @param guard the guard
	 */
	public void addGuard(Guard guard) {
		GuardAttribute ga = getGuardAttribute(guard.getValue().toString(), guard);
		ga.addGuard(guard);
	}
	
	/**
	 * Removes the guard of the transition.
	 *
	 * @param guard the guard
	 */
	public void removeGuard(Guard guard) {
		GuardAttribute ga = getGuardAttribute(guard.getValue().toString(), guard);
		ga.removeGuard(guard);
	}
	
	/**
	 * Removes the all guards of the transition.
	 */
	public void removeAllGuards() {
		GuardAttribute ga = getGuardAttribute(null, null);
		ga.removeAllGuards();
	}
	
	/**
	 * Gets the guards of the transition.
	 *
	 * @return the guards
	 */
	public List<Guard> getGuards() {
		GuardAttribute ga = getGuardAttribute(null, null);
		return ga.getGuards();
	}
	
	
	/**
	 * Fire the transition.
	 *
	 * @return the transition
	 */
	public Transition fire() {
		for (TransitionListener listener : transitionListener)
			listener.fire(this);
		
		return this;
	}
	
	/**
	 * Checks if the transition is enabled.
	 *
	 * @return true, if is enabled
	 */
	public boolean isEnabled() {
		boolean is = false;
		for (TransitionListener listener : transitionListener)
			is = listener.isEnabled(this);
		
		return is;
	}
			
	/**
	 * Adds the transition listener.
	 *
	 * @param listener the listener
	 */
	public void addTransitionListener(TransitionListener listener) {
		if (!transitionListener.contains(listener))
			transitionListener.add(listener);
	}
	

	/**
	 * Checks if the transition is timed.
	 *
	 * @return true, if is timed
	 */
	public boolean isTimed() {
		return false;
	}
	
	/**
	 * Gets the rate.
	 *
	 * @return the rate
	 */
	public double getRate() {
		return 1.0;
	}
	
	/**
	 * Gets the places.
	 *
	 * @return the places
	 */
	public HashSet<Place> getPlaces() {
		HashSet<Place> places = new HashSet<Place>();
		for (AbstractArc a : inArcs)
			places.add(a.getPlace());
		for (AbstractArc a : outArcs)
			places.add(a.getPlace());
		return places;
	}
	
	/**
	 * Is the transition reversible.
	 *
	 * @return reversibles
	 */
	public boolean isRevTransition(){
		return isRevTransition;
	}
	
	public void setRevTransition(boolean isRev){
		isRevTransition = isRev;
	}
}
