/*************************************************************************************
* The Petri-Net Add-on is (c) 2012 Plant Bioinformatics Group, IPK Gatersleben,
* http://bioinformatics.ipk-gatersleben.de
* The source code for this project, is available under the GPL license v2.0 
* available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html. By using this
* Add-on and VANTED you need to accept the terms and conditions of this
* license. For further details see GPL_license.txt in the root folder of this project.
************************************************************************************/

package vanted.petrinetelements.misc;

import java.util.List;

import org.graffiti.graph.Edge;
import org.graffiti.graph.Graph;
import org.graffiti.graph.Node;

import vanted.Parser;
import vanted.petrinetelements.AbstractArc;
import vanted.petrinetelements.AbstractToken;
import vanted.petrinetelements.Place;
import vanted.petrinetelements.TokenContinuous;
import vanted.petrinetelements.TokenDiscrete;
import vanted.petrinetelements.cpn.TokenObject;

// TODO: Auto-generated Javadoc
/**
 * The Enum PetriNetType.
 */
public enum PetriNetType {	
	/** The DISCRETE Petri net type. */
	DISCRETE("discrete"), 

	/** The CONTINUOUS Petri net type. */
	CONTINUOUS("continuous"), 
 
	/** The HYBRID Petri net type. */
	HYBRID("hybrid"), 
 
	/** The OBJECT Petri net type. */
	OBJECT("object"), 
 
	/** The EMPTY Petri net type. */
	EMPTY("empty");	
	
	private String name;
	
	PetriNetType(String name) {
		this.name = name;
	}	
	
	public String getName(){
		return name;
	}
	
	public static PetriNetType fromName(String name){
		if(name==null)return null;
		
		for(PetriNetType type : PetriNetType.values())
			if(name.equalsIgnoreCase(type.name))
				return type;
		
		return null;
	}
	
	/**
	 * Gets the type of a token.
	 *
	 * @param places the places
	 * @return the Petri net type
	 */
	public static PetriNetType getTokenType(List<Place> places){
		boolean discrete=false;
		boolean continuous=false;		
		for(Place place : places){
			for (AbstractToken token : place.getTokens()) 
				if (token instanceof TokenDiscrete)
					discrete = true;
				else if (token instanceof TokenContinuous)
					continuous = true;
				else if (token instanceof TokenObject)
					return OBJECT;
		}
		
		if(!discrete && !continuous)
			return EMPTY;
		
		if(discrete && continuous)
			return HYBRID;
		
		if(discrete)
			return DISCRETE;
		
		if (continuous)
			return CONTINUOUS;		
			
		return OBJECT;
	}
	
	
	/**
	 * Gets the type of a Petri net from the places. Checks the tokens, capacities and arc-weights. 
	 *
	 * @param places the places
	 * @return the Petri net type
	 */
	public static PetriNetType getType(List<Place> places){
		boolean discrete=false;
		boolean continuous=false;		
		for(Place place : places){
			for (AbstractToken token : place.getTokens()) 
				if (token instanceof TokenDiscrete)
					discrete = true;
				else if (token instanceof TokenContinuous)
					continuous = true;
				else if (token instanceof TokenObject)
					return OBJECT;
		}
		
		for(Place place : places){
			if(Parser.parseAttr(place.getCapacity().getValue().toString()) instanceof Double)
				if(Double.parseDouble(place.getCapacity().getValue().toString()) != Integer.MAX_VALUE)
					continuous=true;
		}
		
		for(Place place : places){
			for(AbstractArc arc : place.getInArcs()){
				if(Parser.parseAttr(arc.getArcWeight().getValue().toString()) instanceof Double)
					if(Double.parseDouble(arc.getArcWeight().getValue().toString()) != 1)
						continuous=true;
			}
			for(AbstractArc arc : place.getOutArcs()){
				if(Parser.parseAttr(arc.getArcWeight().getValue().toString()) instanceof Double)
					if(Double.parseDouble(arc.getArcWeight().getValue().toString()) != 1)
						continuous=true;
			}				
		}
		
		if(!discrete && !continuous)
			return EMPTY;
		
		if(discrete && continuous)
			return HYBRID;
		
		if(discrete)
			return DISCRETE;
		
		if (continuous)
			return CONTINUOUS;		
			
		return OBJECT;
	}
	
	/**
	 * Gets the type from a graph. Checks the tokens, capacities and arc-weights.
	 *
	 * @param graph the graph
	 * @return the Petri net type
	 */
	public static PetriNetType getType2(Graph graph){
		boolean discrete=false;
		boolean continuous=false;		
		for(Node node : graph.getNodes()){
			if(Place.isPlace(node)){
				Place place = new Place(node);
				for (AbstractToken token : place.getTokens()) {
					if (token instanceof TokenDiscrete)
						discrete = true;
					else if (token instanceof TokenContinuous)
						continuous = true;
					else if (token instanceof TokenObject)
						return OBJECT;
				}

				if(Parser.parseAttr(place.getCapacity().getValue().toString()) instanceof Double)
					if(Double.parseDouble(place.getCapacity().getValue().toString()) != Integer.MAX_VALUE)
						continuous=true;				
			}			
		}
		
		for(Edge edge : graph.getEdges()){
			if(edge.isDirected()){
				AbstractArc arc = new AbstractArc(edge) {};
				if(Parser.parseAttr(arc.getArcWeight().getValue().toString()) instanceof Double)
					if(Double.parseDouble(arc.getArcWeight().getValue().toString()) != 1)
						continuous=true;
			}
		}
		
		if(!discrete && !continuous)
			return EMPTY;
		
		if(discrete && continuous)
			return HYBRID;
		
		if(discrete)
			return DISCRETE;
		
		if (continuous)
			return CONTINUOUS;		
			
		return OBJECT;
	}	
}
