/*************************************************************************************
* The Petri-Net Add-on is (c) 2012 Plant Bioinformatics Group, IPK Gatersleben,
* http://bioinformatics.ipk-gatersleben.de
* The source code for this project, is available under the GPL license v2.0 
* available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html. By using this
* Add-on and VANTED you need to accept the terms and conditions of this
* license. For further details see GPL_license.txt in the root folder of this project.
************************************************************************************/

package vanted.validation;

import java.util.ArrayList;
import java.util.List;

import org.AttributeHelper;
import org.graffiti.editor.MainFrame;
import org.graffiti.graph.Edge;
import org.graffiti.graph.Graph;
import org.graffiti.graph.Node;
import org.graffiti.selection.Selection;
import org.graffiti.selection.SelectionModel;

import vanted.petrinetelements.NodePN;
import vanted.petrinetelements.Place;
import vanted.petrinetelements.Transition;

// TODO: Auto-generated Javadoc
/**
 * The Class Validator controls the Petri net structure for validity.
 *
 * @author Kevin Pucknat
 */
public class Validator {
	
	/** The transition. */
	private MyNodesArray transition = new MyNodesArray();
	
	/** The place. */
	private MyNodesArray place = new MyNodesArray();
	
	/** The error nodes. */
	private List<Node> errorNodes = new ArrayList<Node>();
	
	/** The error edges. */
	private List<Edge> errorEdges = new ArrayList<Edge>();
	
	/** The nodes. */
	private List<Node> nodes = new ArrayList<Node>();
	
	/** The hidden nodes. */
	private int hiddenNodes;
		
	/**
	 * Checks if the Petri net is correct.
	 *
	 * @param graph the graph
	 * @return true, if is pN correct
	 */
	public boolean isPNCorrect(Graph graph){		
		return isCorrect(graph, true);
	}
	
	/**
	 * Checks if the Petri net is correct in silent mode (silent = no message if the Petri net is correct).
	 *
	 * @param graph the graph
	 * @return true, if is pN correct silent
	 */
	public boolean isPNCorrectSilent(Graph graph){
		return isCorrect(graph, false);
	}
	
	/**
	 * Checks if is correct.
	 *
	 * @param graph the graph
	 * @param showOK the show ok
	 * @return true, if is correct
	 */
	private boolean isCorrect(Graph graph, boolean showOK){
		if(graph==null)return false;
		if(graph.getGraphElements().size()==0)return true;
				
		hiddenNodes=0;
		
		for(Node node : graph.getNodes()){
			NodePN nodePN = new NodePN(node);
			if(!nodePN.getIgnoreNode().getValue())
				nodes.add(node);		
			else
				hiddenNodes++;
		}
		
		errorNodes.clear();
		errorEdges.clear();
		
		boolean error = false;
		
		if(findSimpleErrors(graph)) return false;
		
		for(Node node : graph.getNodes())
			if(node.getEdges().size()==0){
				if(new NodePN(node).getIgnoreNode().getValue())continue;
				MainFrame.showMessageDialog("The Petri-Net has at least one single node", "Petri-Net validation");
				errorNodes.add(node);
				error=true;
			}
				
		for(Edge edge : graph.getEdges()){	
			if(isVisuallyInvertedEdge(edge)){
				error=true;
				errorEdges.add(edge);
			}
			
			if(!nodes.contains(edge.getSource()))continue;
			if(!nodes.contains(edge.getTarget()))continue;
			if(Transition.isTransition(edge.getSource()) && Transition.isTransition(edge.getTarget())){	//Transitions			
				error = true;
				
				if(!transition.contains(edge.getSource(), edge.getTarget() )){
					if(!errorNodes.contains(edge.getSource()))
						errorNodes.add(edge.getSource());
					
					if(!errorNodes.contains(edge.getTarget()))
						errorNodes.add(edge.getTarget());
					
					transition.add(new MyNodes(edge.getSource(), edge.getTarget()));
				}
			}

			if(Place.isPlace(edge.getSource()) && Place.isPlace(edge.getTarget())){	//Places
				error = true;

				if(!place.contains(edge.getSource(), edge.getTarget() )){
					if(!errorNodes.contains(edge.getSource()))
						errorNodes.add(edge.getSource());
					
					if(!errorNodes.contains(edge.getTarget()))
						errorNodes.add(edge.getTarget());
				
					place.add(new MyNodes(edge.getSource(), edge.getTarget()));
				}
			}
		}

		if(error){
			SelectionModel sm = MainFrame.getInstance().getActiveEditorSession().getSelectionModel();
			Selection selection = new Selection("errors");
			if(errorNodes.size()>0)
				selection.addAll(errorNodes);
			
			if(errorEdges.size()>0){
				selection.addAll(errorEdges);
				MainFrame.showMessageDialog("<html>The Petri-Net has "+errorEdges.size()+" visually inverted arcs/edges.<br>", "Petri-Net validation");
			}
			
			
	        sm.setActiveSelection(selection);
	        sm.selectionChanged();
		}
		
		return errors(error, showOK);
	}
	
	/**
	 * Checks if edges are visually inverted.
	 *
	 * @param ed the ed
	 * @return true, if is visually inverted edge
	 */
	public static boolean isVisuallyInvertedEdge(Edge ed) {
        String head = AttributeHelper.getArrowhead(ed);
        String tail = AttributeHelper.getArrowtail(ed);
        return (head == null || head.length() <= 0) && tail != null && tail.length() > 0;
    }
	
	/**
	 * Validity message.
	 *
	 * @param error the error
	 * @param showOK the show ok
	 * @return true, if successful
	 */
	private boolean errors(boolean error, boolean showOK){
		if(!error){
			if(showOK)
				MainFrame.showMessageDialog("<html>The Petri-Net is correct!", "Petri-Net validation");
			return true;
		}else{
			MainFrame.showMessageDialog("<html>The Petri-Net contains errors!", "Petri-Net validation");
			return false;
		}
	}
	
	/**
	 * Find simple errors: undirected edges, no edges, only one node
	 *
	 * @param graph the graph
	 * @return true, if successful
	 */
	private boolean findSimpleErrors(Graph graph){		
		if(graph.getNumberOfUndirectedEdges()>0){
			MainFrame.showMessageDialog("Error: a PN has only directed arcs, containing undirected arcs: "+graph.getNumberOfUndirectedEdges(), "Petri-Net validation");
			return true;
		}
		
		if(graph.getEdges().size()==0){
			MainFrame.showMessageDialog("The Petri-Net contains no arc", "Petri-Net validation");
			return true;
		}
			
		if(graph.getNodes().size()-hiddenNodes==1){
			MainFrame.showMessageDialog("The Petri-Net contains only one node", "Petri-Net validation");
			return true;
		}	
			
		return false;
	}	
}
