/*************************************************************************************
* This file is based on the Pipe2 project (http://pipe2.sourceforge.net/).
* The source code for this project is available under the OSL license 3.0 available at
* http://www.opensource.org/licenses/osl-3.0.php. By using this file you need to accept
* the terms and conditions of this license. For further details see OSL_license.txt 
* in the root folder of this project.
************************************************************************************/

package pipe.analysis.invariants;

/**
 * Invariant analysis module.
 * 
 * @author Manos Papantoniou
 * @author Michael Camacho
 * @author Maxim (GUI and some cleanup)
 */

import java.util.Date;
import java.util.List;

import pipe.analysis.Matrix;
import de.ipk_gatersleben.ag_nw.graffiti.services.task.BackgroundTaskStatusProviderSupportingExternalCallImpl;

//This file is based on the Pipe2 project (http://pipe2.sourceforge.net/).

public class InvariantAnalysis
{
	
	/**
	 * This class has been modified to include a reference to the Petri Net
	 * being analysed so that methods other Run() can access data more easily.
	 * Nadeem 28/05/2005
	 */

	private Matrix _incidenceMatrix;
	private Matrix _PInvariants;
	private Matrix _TInvariants;
	
	private String pInvariant = "";
	private String tInvariant = "";
	private List<String> names;

	
	/**
	 * Invariant analysis was originally only performed on the initial markup
	 * matrix. This is not what the user expects however as when changes are made
	 * to the petri net, the invariant analysis then does not change to reflect
	 * this. The method calls have been changed to pass the current markup matrix
	 * as the parameter for invariant analysis.
	 * 
	 * @param pn
	 * @author Nadeem Akharware
	 * @version 1.2
	 * @return
	 */
	public String startAnalyse(double[] currentMarking, double[][] incidenceMatrix, BackgroundTaskStatusProviderSupportingExternalCallImpl status, List<String> names)
	{
		this.names = names;
		Date start_time = new Date(); // start timer for program execution

		if (incidenceMatrix.length == 0)
		{
			return "";
		}
		_incidenceMatrix = new Matrix(incidenceMatrix);
		
		String output = findNetInvariants(currentMarking, status); // Nadeem 26/05/2005
		
		Date stop_time = new Date();
		double etime = (stop_time.getTime() - start_time.getTime()) / 1000.;
		return output + "<br>Analysis time: " + etime + "s";
	}
	
	/**
	 * Find the net invariants.
	 * 
	 * @param M
	 *           An array containing the current marking of the net.
	 * @return A string containing the resulting matrices of P and T
	 *         Invariants or "None" in place of one of the matrices if it does
	 *         not exist.
	 */
	String findNetInvariants(double[] M, BackgroundTaskStatusProviderSupportingExternalCallImpl status)
	{
		return reportPInvariants(M, status) + "<br>" + reportTInvariants(M, status) + "<br>";
	}
	
	/**
	 * Reports on the P invariants.
	 * 
	 * @param M
	 *           An array containing the current marking of the net.
	 * @return A string containing the resulting matrix of P Invariants,
	 *         the P equations and some analysis
	 */
	String reportPInvariants(double[] M, BackgroundTaskStatusProviderSupportingExternalCallImpl status)
	{
		_PInvariants = findVectors(_incidenceMatrix.transpose(), status);
		
		if (_PInvariants == null) {
			pInvariant = "Invariant analysis was canceled by user";
			return "Invariant analysis was canceled by user";
		}
		
		String result = "<h3>P-Invariants</h3>";
		
		if (_PInvariants.isCovered())
		{
			result += "The net is covered by positive P-Invariants, " +
					"therefore it is bounded.";
		}
		else
		{
			result += "The net is not covered by positive P-Invariants, " +
					"therefore we do not know if it is bounded.";
		}
		pInvariant = result + "<br>" + findPEquations(M);
		return pInvariant;
	}
	
	/**
	 * Reports on the T invariants.
	 * 
	 * @param M
	 *           An array containing the current marking of the net.
	 * @return A string containing the resulting matrix of T Invariants and
	 *         some analysis of it
	 */
	String reportTInvariants(double[] M, BackgroundTaskStatusProviderSupportingExternalCallImpl status)
	{
		_TInvariants = findVectors(_incidenceMatrix, status);
		
		if (_TInvariants == null) {
			tInvariant = "Invariant analysis was canceled by user";
			return "";
		}
		
		String result = "<h3>T-Invariants</h3>";
		
		if (_TInvariants.isCovered())
		{
			result += "The net is covered by positive T-Invariants, " +
					"therefore it might be bounded and live.";
		}
		else
		{
			result += "The net is not covered by positive T-Invariants, " +
					"therefore we do not know if it is bounded and live.";
		}
		tInvariant = result + "<br>";
		return tInvariant;
	}
	
	// <Marc>
	/*
	 * It returns a PNMatrix containing the place invariants of the sourceDataLayer net.
	 * @param sourceDataLayer A _dataLayer type object with all the information about
	 * the petri net.
	 * @return a PNMatrix where each column contains a place invariant.
	 */
	public Matrix getPInvariants(double[][] incidenceMatrix, BackgroundTaskStatusProviderSupportingExternalCallImpl status)
	{
		if (incidenceMatrix.length == 0)
		{
			return null;
		}
		_incidenceMatrix = new Matrix(incidenceMatrix);
		
		return findVectors(_incidenceMatrix.transpose(), status);
	}
	
	/*
	 * It returns a PNMatrix containing the transition invariants of the sourceDataLayer net.
	 * @param sourceDataLayer A _dataLayer type object with all the information about
	 * the petri net.
	 * @return a PNMatrix where each column contains a transition invariant.
	 */
	public Matrix getTInvariants(double[][] incidenceMatrix, BackgroundTaskStatusProviderSupportingExternalCallImpl status)
	{
		if (incidenceMatrix.length == 0)
		{
			return null;
		}
		_incidenceMatrix = new Matrix(incidenceMatrix);
		
		return findVectors(_incidenceMatrix, status);
	}
	
	// </Marc>
	
	/**
	 * Find the P equations of the net.
	 * 
	 * @param currentMarking
	 *           An array containing the current marking of the net.
	 * @return A string containing the resulting P equations,
	 *         empty string if the equations do not exist.
	 */
	String findPEquations(double[] currentMarking)
	{
		if(names==null)return "";
		String eq = "<h3>P-Invariant equations</h3>";
		int m = _PInvariants.getRowDimension();
		int n = _PInvariants.getColumnDimension();
		if (n < 1)
		{ // if there are no P-invariants don't return any equations
			return "";
		}
		
		Matrix col = new Matrix(m, 1);
		int rhs = 0; // the right hand side of the equations
		
		// form the column vector of current marking
		Matrix M = new Matrix(currentMarking, currentMarking.length);
		// for each column c in PInvariants, form one equation of the form
		// a1*p1+a2*p2+...+am*pm = c.transpose*M where a1, a2, .., am are the
		// elements of c
		for (int i = 0; i < n; i++)
		{ // column index
			for (int j = 0; j < m; j++)
			{ // row index
				// form left hand side:
				double a = _PInvariants.get(j, i);
				if (a > 1)
				{
					eq += Double.toString(a);
				}
				if (a > 0)
				{
					eq += "M(" + names.get(j) + ") + "; // Nadeem 28/05/2005
				}
			}
			// replace the last occurance of "+ "
			eq = eq.substring(0, (eq.length() - 2)) + "= ";
			
			// form right hand side
			col = _PInvariants.getMatrix(0, m - 1, i, i);
			rhs = col.transpose().vectorTimes(M);
			eq += rhs + "<br>"; // and separate the equations
		}
		return eq;
	}
	
	/**
	 * Transform a matrix to obtain the minimal generating set of vectors.
	 * 
	 * @param c
	 *           The matrix to transform.
	 * @return A matrix containing the vectors.
	 */
	public Matrix findVectors(Matrix c, BackgroundTaskStatusProviderSupportingExternalCallImpl status)
	{
		/*
		 * | Tests Invariant Analysis IModule
		 * |
		 * | C = incidence matrix.
		 * | B = identity matrix with same number of columns as C.
		 * | Becomes the matrix of vectors in the end.
		 * | pPlus = integer array of +ve indices of a row.
		 * | pMinus = integer array of -ve indices of a row.
		 * | pPlusMinus = set union of the above integer arrays.
		 */
		int m = c.getRowDimension(), n = c.getColumnDimension();
		
		// generate the nxn identity matrix
		Matrix B = Matrix.identity(n, n);
				
		// arrays containing the indices of +ve and -ve elements in a row vector
		// respectively
		int[] pPlus, pMinus;
		
		// while there are no zero elements in C do the steps of phase 1
		// --------------------------------------------------------------------------------------
		// PHASE 1:
		// --------------------------------------------------------------------------------------
		while (!(c.isZeroMatrix()))
		{
			if (status.wantsToStop())
				return null;
			if (c.checkCase11())
			{
				// check each row (case 1.1)
				for (int i = 0; i < m; i++)
				{
					pPlus = c.getPositiveIndices(i); // get +ve indices of ith row
					pMinus = c.getNegativeIndices(i); // get -ve indices of ith row
					if (isEmptySet(pPlus) || isEmptySet(pMinus))
					{ // case-action 1.1.a
						// this has to be done for all elements in the union pPlus U pMinus
						// so first construct the union
						int[] pPlusMinus = uniteSets(pPlus, pMinus);
						
						// eliminate each column corresponding to nonzero elements in pPlusMinus union
						for (int j = pPlusMinus.length - 1; j >= 0; j--)
						{
							if (status.wantsToStop())
								return null;
							if (pPlusMinus[j] != 0)
							{
								c = c.eliminateCol(pPlusMinus[j] - 1);								
								B = B.eliminateCol(pPlusMinus[j] - 1);
//								n--; // reduce the number of columns since new matrix is smaller
							}
						}
					}
					resetArray(pPlus); // reset pPlus and pMinus to 0
					resetArray(pMinus);
				}
			}
			else{//1.1.b
				if (c.cardinalityCondition() >= 0)
				{
					phase11b1(c, B, status);
				}
				else
				{
					phase11b2(c, B, status);
				}//end 1.1.b.2
			}//end 1.1.b
		}
		// System.out.println("end of phase one");
		// END OF PHASE ONE, now B contains a pseudodiagonal positive basis of Ker C
		// --------------------------------------------------------------------------------------
		// PHASE 2:
		// --------------------------------------------------------------------------------------
		// h is -1 at this point, make it equal to the row index that has a -ve element.
		// rowWithNegativeElement with return -1 if there is no such row, and we exit the loop.
		B = phase2(B, status);
		
		// delete from B all cols having non minimal support
		// k is the index of column to be eliminated, if it is -1 then there is
		// no col to be eliminated
		int k = 0;
		// form a matrix with columns the row indices of non-zero elements
//		Matrix bi = B.nonZeroIndices();
//		
//		while (k > -1)
//		{
//			k = bi.findNonMinimal();			
//			if (k != -1)
//			{
//				System.out.println("entry with non minimal support!!!!!!!!!!!!!!!");
//				B = B.eliminateCol(k);
//				bi = B.nonZeroIndices();
//			}
//		}
		
		// display the result
		// System.out.println("Minimal generating set (after phase 2):");
		// B.print(2, 0);
					
		checkGCD(B);
		
//		System.out.println("final");
//		printMatrix(B);
		return B;
	}
	
	private void phase11b1(Matrix c, Matrix B, BackgroundTaskStatusProviderSupportingExternalCallImpl status) {
//		System.out.println("1.1.b.1");
		while (c.cardinalityCondition() >= 0)
		{
			if (status.wantsToStop())
				return;
			// while there is a row in the C matrix that satisfies the cardinality condition
			// do a linear combination of the appropriate columns and eliminate the appropriate column.
			int cardRow = -1; // the row index where cardinality == 1
			cardRow = c.cardinalityCondition();
			// get the column index of the column to be eliminated
			int k = c.cardinalityOne();
			if (k == -1)
			{
				System.out.println("Error");
			}
			
			// get the comlumn indices to be changed by linear combination
			int j[] = c.colsToUpdate();
			
			// update columns with linear combinations in matrices C and B
			// first retrieve the coefficients
			double[] jCoef = new double[c.getRowDimension()];
			for (int i = 0; i < j.length; i++)
			{
				if (j[i] != 0)
				{
					jCoef[i] = Math.abs(c.get(cardRow, (j[i] - 1)));
				}
			}
			
			// do the linear combination for C and B
			// k is the column to add, j is the array of cols to add to
			c.linearlyCombine(k, Math.abs(c.get(cardRow, k)), j, jCoef);			
			B.linearlyCombine(k, Math.abs(c.get(cardRow, k)), j, jCoef);
			
			// eliminate column of cardinality == 1 in matrices C and B
			c = c.eliminateCol(k);
			
//			System.out.println("eliminate column of cardinality == 1 in matrices C and B");
//			System.out.println("1.1.b.1");
//			printMatrix(c);
			B = B.eliminateCol(k);
			// reduce the number of columns since new matrix is smaller
//			n--;
			break;
		}
		
	}

	private void phase11b2(Matrix c, Matrix B, BackgroundTaskStatusProviderSupportingExternalCallImpl status){
	// row annihilations (condition 1.1.b.2)
		// operate only on non-zero rows of C (row index h)
		// find index of first non-zero row of C (int h)
//		System.out.println("1.1.b.2");
		int h = c.firstNonZeroRowIndex();
		while ((h = c.firstNonZeroRowIndex()) > -1)
		{
			if (status.wantsToStop())
				return;
			// the column index of the first non zero element of row h
			int k = c.firstNonZeroElementIndex(h);
			
			// find first non-zero element at column k, chk
			double chk = c.get(h, k);
			
			// find all the other indices of non-zero elements in that row chj[]
			int[] chj = new int[c.getColumnDimension()-1];
			chj = c.findRemainingNZIndices(h);
			
			while (!(isEmptySet(chj)))
			{
				if (status.wantsToStop())
					return;
				// chj empty only when there is just one nonzero element in the
				// whole row, this should not happen as this case is eliminated
				// in the first step, so we would never arrive at this while()
				// with just one nonzero element
				
				// find all the corresponding elements in that row (coefficients jCoef[])
				double[] jCoef = c.findRemainingNZCoef(h);
				
				// adjust linear combination coefficients according to sign
				double[] alpha, beta; // adjusted coefficients for kth and remaining columns respectively
//				System.out.println("alpha");
				alpha = alphaCoef(chk, jCoef);
//				printMatrix(new Matrix(alpha, 1));
				beta = betaCoef(chk, jCoef.length);
//				System.out.println("beta");
//				printMatrix(new Matrix(beta, 1));
				
				// linearly combine kth column, coefficient alpha, to jth columns, coefficients beta
				c.linearlyCombine(k, alpha, chj, beta);
				B.linearlyCombine(k, alpha, chj, beta);
				
				// delete kth column
				c = c.eliminateCol(k);
//				System.out.println("1.1.b.2");
//				printMatrix(c);
				B = B.eliminateCol(k);
				
										
				chj = c.findRemainingNZIndices(h);
				break;
			}
			break;
		}
		// show the result
		// System.out.println("Pseudodiagonal positive basis of Ker C after phase 1:");
		// B.print(2, 0);
	}
			
	private Matrix phase2(Matrix B, BackgroundTaskStatusProviderSupportingExternalCallImpl status){
		int h;
//		System.out.println("phase 2:");
//		printMatrix(B);
		int[] pPlus = null;
		int[] pMinus = null;
		int m=0;		
		int n=0;
		while ((h = B.rowWithNegativeElement()) > -1)
		{
			if (status.wantsToStop())
				return null;
			pPlus = B.getPositiveIndices(h); // get +ve indices of hth row (1st col = index 1)
			pMinus = B.getNegativeIndices(h); // get -ve indices of hth row (1st col = index 1)
			
			// effective length is the number of non-zero elements
			int pPlusLength = effectiveSetLength(pPlus);
			int pMinusLength = effectiveSetLength(pMinus);
			
			if (pPlusLength != 0)
			{ // set of positive coef. indices must me non-empty
				// form the cross product of pPlus and pMinus
				// for each pair (j, k) in the cross product, operate a linear combination on the columns
				// of indices j, k, in order to get a new col with a zero at the hth element
				// The number of new cols obtained = the number of pairs (j, k)
				for (int j = 0; j < pPlusLength; j++)
				{
					for (int k = 0; k < pMinusLength; k++)
					{
						// coefficients of row h, cols indexed j, k in pPlus, pMinus
						// respectively
						int jC = pPlus[j] - 1, kC = pMinus[k] - 1;
						
						// find coeficients for linear combination, just the abs values
						// of elements this is more efficient than finding the least
						// common multiple and it does not matter since later we will
						// find gcd of col and we will normalise with that the col
						// elements
						double a = -B.get(h, kC), b = B.get(h, jC);
						
						// create the linear combination a*jC-column + b*kC-column, an
						// IntMatrix mx1 where m = number of rows of B
						m = B.getRowDimension();
						Matrix v1 = new Matrix(m, 1); // column vector mx1 of zeros
						Matrix v2 = new Matrix(m, 1); // column vector mx1 of zeros
						v1 = B.getMatrix(0, m - 1, jC, jC);
						v2 = B.getMatrix(0, m - 1, kC, kC);
						v1.timesEquals(a);
						v2.timesEquals(b);
						v2.plusEquals(v1);
						
						// find the gcd of the elements in this new col
						double V2gcd = v2.gcd();
						
						// divide all the col elements by their gcd if gcd > 1
						if (V2gcd > 1)
						{
							v2.divideEquals(V2gcd);
						}
						
						// append the new col to B
						n = B.getColumnDimension();
						Matrix f = new Matrix(m, n + 1);
						f = B.appendVector(v2);
						B = f.copy();
					}
				} // endfor (j,k) operations
				
				// delete from B all cols with index in pMinus
				for (int ww = 0; ww < pMinusLength; ww++)
				{
					B = B.eliminateCol(pMinus[ww] - 1);
				}
				
			} // endif
		} // end while
		return B;
	}
	
	private double alpha(double chk, double chj){
		//sign(Chj)!=sign(Chk)
		if(chj<0 && chk>=0 || chj>0 && chk<=0)
			return Math.abs(chj);
		else
			return -Math.abs(chj);		
	}
	
	private double beta(double chk, double chj){
		//sign(Chj)!=sign(Chk)
		if(chj<0 && chk>=0)
			return Math.abs(chk);
		else
			return Math.abs(chk);
	}
	
	//get minimal invariants
	private void checkGCD(Matrix B){
		Matrix temp;
		for(int j=0; j<B.getColumnDimension();j++){
			temp = B.getMatrix(0, B.getRowDimension()-1, j, j);
			
			double Tempgcd = temp.gcd();
			
			// divide all the col elements by their gcd if gcd > 1
			if (Tempgcd > 1)
			{
				temp.divideEquals(Tempgcd);				
				for(int i=0; i<temp.getRowDimension();i++){					
					B.set(i, j, temp.get(i, 0));
				}
			}
		}	
	}
	
	
	/**
	 * find the number of non-zero elements in a set
	 * 
	 * @param pSet
	 *           The set count the number of non-zero elements.
	 * @return The number of non-zero elements.
	 */
	private int effectiveSetLength(int[] pSet)
	{
		int effectiveLength = 0; // number of non-zero elements
		int setLength = pSet.length;
		
		for (int i = 0; i < setLength; i++)
		{
			if (pSet[i] != 0)
			{
				effectiveLength++;
			}
			else
			{
				return effectiveLength;
			}
		}
		return effectiveLength;
	}
	
	/**
	 * adjust linear combination coefficients according to sign
	 * if sign(j) <> sign(k) then alpha = abs(j) beta = abs(k)
	 * if sign(j) == sign(k) then alpha = -abs(j) beta = abs(k)
	 * 
	 * @param k
	 *           The column index of the first coefficient
	 * @param j
	 *           The column indices of the remaining coefficients
	 * @return The adjusted alpha coefficients
	 */
	private double[] alphaCoef(double k, double[] j)
	{
		int n = j.length; // the length of one row
		double[] alpha = new double[n];
		
		for (int i = 0; i < n; i++)
		{
			if ((k * j[i]) < 0)
			{
				alpha[i] = Math.abs(j[i]);
			}
			else
			{
				alpha[i] = -Math.abs(j[i]);
			}
		}
		return alpha;
	}
	
			
	/**
	 * adjust linear combination coefficients according to sign
	 * if sign(j) <> sign(k) then alpha = abs(j) beta = abs(k)
	 * if sign(j) == sign(k) then alpha = -abs(j) beta = abs(k)
	 * 
	 * @param chk
	 *           The first coefficient
	 * @param n
	 *           The length of one row
	 * @return The adjusted beta coefficients
	 */
	private double[] betaCoef(double chk, int n)
	{
		double[] beta = new double[n];
		double abschk = Math.abs(chk);
		
		for (int i = 0; i < n; i++)
		{
			beta[i] = abschk;
		}
		return beta;
	}
	
	private void resetArray(int[] a)
	{
		for (int i = 0; i < a.length; i++)
		{
			a[i] = 0;
		}
	}
	
	/**
	 * Unite two sets (arrays of integers) so that if there is a common entry in
	 * the arrays it appears only once, and all the entries of each array appear
	 * in the union. The resulting array size is the same as the 2 arrays and
	 * they are both equal. We are only interested in non-zero elements. One of
	 * the 2 input arrays is always full of zeros.
	 * 
	 * @param A
	 *           The first set to unite.
	 * @param B
	 *           The second set to unite.
	 * @return The union of the two input sets.
	 */
	private int[] uniteSets(int[] A, int[] B)
	{
		int[] union = new int[A.length];
		
		if (isEmptySet(A))
		{
			union = B;
		}
		else
		{
			union = A;
		}
		return union;
	}
	
	/**
	 * check if an array is empty (only zeros)
	 * 
	 * @param pSet
	 *           The set to check if it is empty.
	 * @return True if the set is empty.
	 */
	private boolean isEmptySet(int[] pSet)
	{
		int setLength = pSet.length;
		
		for (int i = 0; i < setLength; i++)
		{
			if (pSet[i] != 0)
			{
				return false;
			}
		}
		return true;
	}
	
	
	public String getPInvariantStr() {
		return pInvariant;
	}
	
	public String getTInvariantStr() {
		return tInvariant;
	}
	
	public Matrix get_PInvariants() {
		return _PInvariants;
	}
	
	public Matrix get_TInvariants() {
		return _TInvariants;
	}
	
	//debug
	private void printMatrix(Matrix matrix){
		for(int n=0; n<matrix.getRowDimension(); n++){
			for(int m=0; m<matrix.getColumnDimension(); m++){
				System.out.print(matrix.get(n, m)+"\t");
			}
			System.out.println();
		}
		System.out.println();
	}
}
