/*************************************************************************************
* This file is based on the Pipe2 project (http://pipe2.sourceforge.net/).
* The source code for this project is available under the OSL license 3.0 available at
* http://www.opensource.org/licenses/osl-3.0.php. By using this file you need to accept
* the terms and conditions of this license. For further details see OSL_license.txt 
* in the root folder of this project.
************************************************************************************/

package pipe.analysis.reachability;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.List;
import org.AttributeHelper;
import org.graffiti.editor.MainFrame;
import org.graffiti.event.ListenerManager;
import org.graffiti.graph.AdjListGraph;
import org.graffiti.graph.Edge;
import org.graffiti.graph.Graph;
import org.graffiti.graph.Node;

import pipe.analysis.Matrix;


import de.ipk_gatersleben.ag_nw.graffiti.services.task.BackgroundTaskStatusProviderSupportingExternalCallImpl;

/**
 * @author Matthew Worthington / Edwin Chung / Will Master
 *         Created module to produce the reachability graph representation of a Petri
 *         net. This module makes use of modifications that were made to the state space
 *         generator to produce a list of possible states (both tangible and non
 *         tangible) which are then transformed into a dot file.
 *         The file is then translated into its graphical layout using
 *         www.research.att.com hosting of Graphviz. It should therefore be noted that
 *         a live internet connection is required. (Feb/March,2007)
 */

//This file is based on the Pipe2 project (http://pipe2.sourceforge.net/).

public class ReachabilityGraphGenerator {
	private AdjListGraph reachabilityGraph;
	private List<Node> nodeList;
	private List<String> markingList;
	public static final String reachabilityGraphName = "reachabilityGraph";
	
	public Graph startAnalyse(Matrix incidenceMatrix, double[][] fMatrix, double[][] bMatrix, double[] capacityVector, double[] marking,
			BackgroundTaskStatusProviderSupportingExternalCallImpl status, int reachabilityMaxNodes, File reachabilityGraphFile, File intermediate, List<String[]> labelidentity) {
		// This will be used to store the reachability graph data
		
		try {						
			boolean generateCoverability = false;
			try {
				StateSpaceGenerator.generate(incidenceMatrix, fMatrix, bMatrix, capacityVector, marking, reachabilityGraphFile, status, reachabilityMaxNodes, intermediate);
			} catch (OutOfMemoryError e) {
				// net seems to be unbounded, let's try to generate the
				// coverability graph
				generateCoverability = true;
			}
			
			double[] currentMarking = marking;
			
			// TODO: reachability graph and coverability graph are the same
			// when the net is bounded so we could just generate the
			// coverability graph
			if (generateCoverability) {
				myTree tree = new myTree(incidenceMatrix, fMatrix, bMatrix, capacityVector,
						currentMarking,
						reachabilityGraphFile, status, intermediate);				
			}
			
			System.gc();
			if (status.wantsToStop())
				return null;
			Graph reachabilityGraph = createGraph(reachabilityGraphFile, generateCoverability, labelidentity);
			
			status.setCurrentStatusText1("");
			
			return reachabilityGraph;
		} catch (OutOfMemoryError e) {
			System.gc();
			return null;
		} catch (StackOverflowError e) {
			return null;
		} catch (ImmediateAbortException e) {
			return null;
		}
		// catch(TimelessTrapException e){
		// return;
		// }
		catch (IOException e) {
			return null;
		} catch (TreeTooBigException e) {
			return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		} 
//		finally {
//			if (reachabilityGraphFile.exists()) {				
//				reachabilityGraphFile.delete();
//			}
//		}
	}
	
	private Graph createGraph(File rgFile, boolean coverabilityGraph, List<String[]> labelidentity)
			throws IOException {		
		reachabilityGraph = new AdjListGraph(new ListenerManager());
		nodeList = new ArrayList<Node>();
		markingList = new ArrayList<String>();
		List<Node> loops = new ArrayList<Node>();
		List<String> loopsTransitions = new ArrayList<String>();
		
		ReachabilityGraphFileHeader header = new ReachabilityGraphFileHeader();
		RandomAccessFile reachabilityFile = null;
		try {
			reachabilityFile = new RandomAccessFile(rgFile, "r");
			header.read(reachabilityFile);
		} catch (IncorrectFileFormatException e1) {
			if(reachabilityFile!=null)
				reachabilityFile.close();
			System.err.println("createGraph: " +
					"incorrect file format on state space file");
			return reachabilityGraph;
		} catch (IOException e1) {	
			if(reachabilityFile!=null)
				reachabilityFile.close();
			System.err.println("createGraph: unable to read header file");
			return reachabilityGraph;
		}
		
		if ((header.getNumStates() + header.getNumTransitions()) > 400) {
			reachabilityFile.close();			
			MainFrame.showMessageDialog("The graph is too big to be displayed properly.", "Graph too big error");
			throw new IOException("There are " + header.getNumStates() + " states with "
					+ header.getNumTransitions() + " arcs. c");
		}
		
		String label;
		String marking;
		
		int stateArraySize = header.getStateArraySize();
		StateRecord record = new StateRecord();
		record.read1(stateArraySize, reachabilityFile);
		label = "S0";
		marking = record.getMarkingString();
		if (record.getTangible()) {// immediate transitions
			addNodeWithAttributes(label, marking, Color.RED, "circle");
		}
		else { // timed transitions
			addNodeWithAttributes(label, marking, Color.RED, "circle");
		}
		
		for (int count = 1; count < header.getNumStates(); count++) {// the root is set above, here are set all other nodes
			record.read1(stateArraySize, reachabilityFile);
			label = "S" + count;
			marking = record.getMarkingString();
			if (record.getTangible()) {// immediate transitions
			
				if (marking.contains("-1"))
					addNodeWithAttributes(label, marking, null, "ellipsenodeshape");
				else
					addNodeWithAttributes(label, marking, null, "circle");
			}
			else {// timed transitions
				addNodeWithAttributes(label, marking, null, "circle");
			}
		}
		
		reachabilityFile.seek(header.getOffsetToTransitions());
		int numberTransitions = header.getNumTransitions();
		TransitionRecord transitions = new TransitionRecord();
		for (int transitionCounter = 0; transitionCounter < numberTransitions; transitionCounter++) {
			transitions = new TransitionRecord();
			transitions.read1(reachabilityFile);
			
			int from = transitions.getFromState();
			int to = transitions.getToState();
			if (from != to) {// no loop
				Edge edge = reachabilityGraph.addEdge(nodeList.get(from), nodeList.get(to), true,
						AttributeHelper.getDefaultGraphicsAttributeForEdge(Color.BLACK, Color.BLACK, true));				
//				AttributeHelper.setLabel(edge, AttributeHelper.getLabel(pn.getAllTransitions().get(transitions.getTransitionNo()).getNode(), ""));
//				edge.setID(pn.getAllTransitions().get(transitions.getTransitionNo()).getIdentity().getValue());				
				AttributeHelper.setLabel(edge, labelidentity.get(transitions.getTransitionNo())[0], "");
				edge.setID(Long.parseLong(labelidentity.get(transitions.getTransitionNo())[1]));				
			}
			else {// loop				
				if (loops.contains(nodeList.get(from))) {
					int index = loops.indexOf(nodeList.get(from));
//					loopsTransitions.set(index, loopsTransitions.get(index) + ", " + pn.getAllTransitions().get(transitions.getTransitionNo()).toString());
//					loopsTransitions.set(index, loopsTransitions.get(index) + ", " + labelidentity.get(transitions.getTransitionNo())[1]);
				}
				else {
					loops.add(nodeList.get(from));
//					loopsTransitions.add(pn.getAllTransitions().get(transitions.getTransitionNo()).getIdentity().getValue() + "");
					loopsTransitions.add(labelidentity.get(transitions.getTransitionNo())[1]);
				}
			}
		}
		
		for (int i = 0; i < loops.size(); i++) {
			Edge edge = reachabilityGraph.addEdge(loops.get(i), loops.get(i), true,
					AttributeHelper.getDefaultGraphicsAttributeForEdge(Color.BLACK, Color.BLACK, true));
			AttributeHelper.setLabel(edge, labelidentity.get(transitions.getTransitionNo())[0]);
			edge.setID(Long.parseLong(labelidentity.get(transitions.getTransitionNo())[1]));
		}
				
		reachabilityFile.close();
				
		return reachabilityGraph;
	}
	
	private Node addNodeWithAttributes(String label, String marking, Color fillColor, String shape) {
		Node node = reachabilityGraph.addNode(AttributeHelper.getDefaultGraphicsAttributeForNode(0, 0));
		AttributeHelper.setLabel(node, label);
		if (!shape.isEmpty())
			AttributeHelper.setShape(node, shape);
		if (fillColor != null)
			AttributeHelper.setFillColor(node, fillColor);
		nodeList.add(node);
		markingList.add(marking);
		
//		NodePN nodePN = new NodePN(node);
//		nodePN.addReachability(new Reachability(marking));
		
		return node;
	}
	
	public List<Node> getNodeList(){
		return nodeList;
	}
	
	public List<String> getMarkingList(){
		return markingList;
	}
}
