/*************************************************************************************
* This file is based on the Pipe2 project (http://pipe2.sourceforge.net/).
* The source code for this project is available under the OSL license 3.0 available at
* http://www.opensource.org/licenses/osl-3.0.php. By using this file you need to accept
* the terms and conditions of this license. For further details see OSL_license.txt 
* in the root folder of this project.
************************************************************************************/

/*
 * Created on 29-Jun-2005
 */
//package pipe.calculations;
package pipe.analysis.reachability;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Stack;

import pipe.analysis.Matrix;
import test.PNHelper;

import de.ipk_gatersleben.ag_nw.graffiti.services.task.BackgroundTaskStatusProviderSupportingExternalCallImpl;


/**
 * @author Nadeem
 * @author Will Master - minor changes 02/2007
 * @author Edwin Chung/Matthew Worthington/Will Master - changes to accomodate
 *         requirements of reachability graph module, overloading generate
 *         method to produce both tangible and non tangible states of net.
 * @author Pere Bonet - fixed a bug in generate method caused by a wrong
 *         calculation of the probability of a transition from a vanishing state
 *         to another state
 * 
 *         This class is used to generate the state space of a General
 *         Stochastic Petri-Net model. It uses an algorithm developed by Dr.
 *         William Knottenbelt in his MSc. thesis "Generalised Markovian
 *         Analysis of Timed Transition Systems" June 1996 Dept. of Computer
 *         Science, University of Cape Town. The algorithm eliminates vanishing
 *         states 'on the fly' and creates the infinitessimal generator matrix
 *         'Q' at the same time.
 * 
 *         There is just one public method and it is static. It returns the
 *         state space array and the matrix Q.
 */

//This file is based on the Pipe2 project (http://pipe2.sourceforge.net/).

public class StateSpaceGenerator {

	private static final boolean DEBUG = false;
	private static final int NUMHASHROWS = 46567;

	// Array storing the transitions fired
	private static final Stack transitions = new Stack();
	private static Matrix incidenceMatrix;
	private static double[][] fMatrix;
	private static double[][] bMatrix;
	private static double[] capacityVector;

	/**
	 * generate() This static method generates the statespace from a GSPN It
	 * uses a hashtable so that it can quickly check whether a state has already
	 * been explored.
	 * 
	 * @param pnmlData
     * @param reachGraph
	 * @throws ImmediateAbortException
	 * @throws IOException
	 * @author Nadeem 30/06/2005
     * @throws OutOfMemoryError
	 */
	public static void generate(Matrix incidenceMatrix, double[][] fMatrix, double[][] bMatrix, double[] capacityVector,
			double[] marking, File reachGraph, BackgroundTaskStatusProviderSupportingExternalCallImpl status, int maxNodes, File intermediate)
			throws OutOfMemoryError,
            ImmediateAbortException {
				
		StateSpaceGenerator.incidenceMatrix = incidenceMatrix;
		StateSpaceGenerator.fMatrix = fMatrix;
		StateSpaceGenerator.bMatrix = bMatrix;
		StateSpaceGenerator.capacityVector = capacityVector;

		State currentMarking = new State(marking);
		int statearraysize = currentMarking.getState().length;

		Queue statesQueue = new Queue();// States waiting to be explored
		Stack tansuccessor = new Stack();
		// Objects for temporarily storing states that haven't been identified
		// as
		// tangible or vanishing
		State sprime = null;
		MarkingState currentState = null; // Used in some loops
		MarkingState s = null; // Used in some loops

		// A record of all explored states. The actual states themselves are not
		// stored here. Instead, just their two hashcodes are used to represent
		// them, one as a key to the hashtable row, the other as the entry in a
		// list at each hashtable row.
		LinkedList[] exploredStates = new LinkedList[NUMHASHROWS];

		// This list is used to temporarily store details of the arcs between
		// states on the reachability graph.
		LinkedList localarcs = new LinkedList();

		// Counters used for creating the reachability graph file
		int numStates = 0;
		int numTransitions = 0;
		int numtransitionsfired = 0;

		// Temporary files for storing tangible states and the transitions
		// between
		// them. They are later combined into one file by the createRGFile
		// method.
		RandomAccessFile outputFile;
		RandomAccessFile esoFile;

		try {
			outputFile = new RandomAccessFile(intermediate, "rw");			
			esoFile = new RandomAccessFile(reachGraph, "rw");
			// Write a blank file header as a place holder for later
			ReachabilityGraphFileHeader header = new ReachabilityGraphFileHeader();
			header.write(esoFile);
		} catch (IOException e) {
			
			System.out.println("Could not create intermediate files.");
			return;
		}

		currentState = new MarkingState(currentMarking, numStates, isTangible(currentMarking));
		numStates++;
		statesQueue.enqueue(currentState);
		addExplored(currentState, exploredStates, esoFile, true);

		while (!statesQueue.isEmpty()) {

			s = (MarkingState) statesQueue.dequeue();
			numtransitionsfired += fire(s, tansuccessor);
			while (!tansuccessor.isEmpty()) {
				status.setCurrentStatusText2("explored states: "+exploredStates.length);
				if(status.wantsToStop()){
					status.setCurrentStatusText2("Abort by user");
					return;
				}
				
				sprime = (State) tansuccessor.pop();
				if (!explored(sprime, exploredStates)) {
					currentState = new MarkingState(sprime, numStates, isTangible(sprime));
					numStates++;
					statesQueue.enqueue(currentState);
					addExplored(currentState, exploredStates, esoFile, true);
				} else {
					int id = identifyState(sprime, exploredStates);
					if (id == -1) {
						throw new ImmediateAbortException(
								"Could not identify previously explored tangible state.");
					}
					currentState = new MarkingState(sprime, id);
				}
				numTransitions += transition(currentState, rate(s, sprime), localarcs);
			}
			// Write all the arcs for the reachability graph to file
			writeTransitions(s, localarcs, outputFile, true);
			// Clear the list so can start again with the next set of arcs
			localarcs.clear();
			status.setCurrentStatusText2("fired transitions: "+ numTransitions+ " explored states: "+numStates);
			// REVISAR-LIMITS
			if (numTransitions > maxNodes) {
				throw new OutOfMemoryError("The net generates in excess of "
						+ maxNodes + " states");
			}
		}
		try {			
			outputFile.close();
		} catch (IOException e1) {
			System.err.println("\nCould not close intermediate file.");
		}		
		System.out.println("\nGenerate Ends, " + numStates
				+ " states found with " + numTransitions + " arcs.");
		try {
			createRGFile(intermediate, esoFile, statearraysize, numStates,
					numTransitions, true);
		
			if (DEBUG) {
				/* Write a csv file indicating the Hashtable distribution */
				FileWriter htdist = new FileWriter("HashTableDist.csv");
				for (int row = 0; row < NUMHASHROWS; row++) {
					htdist.write(Integer.toString(row) + ",");
					if (exploredStates[row] == null) {
						htdist.write("0\n");
					} else {
						htdist.write(Integer.toString(exploredStates[row].size())
								+ "\n");
					}
				}
				htdist.close();
				System.out
						.println("Finished writing hashtable distribution to file.");
			}
		}catch (Exception e) {
			e.printStackTrace();
		}finally{
			try {
				outputFile.close();
				esoFile.close();	
			} catch (IOException e) {
				e.printStackTrace();
			}			
		}
	}

	/**
	 * isTangible() Tests whether the state passed as an argument is tangible or
	 * vanishing.
	 * 
	 * @param pnmlData
	 * @param marking
	 * @return
	 */
	private static boolean isTangible(State marking) {
//		List<Transition> trans = pn.getAllTransitions();
//		int numTrans = trans.size();
		boolean hasTimed = false;
		boolean hasImmediate = false;
		
		boolean[] enabledTransitions = PNHelper.getEnabledTransitionsFromMarking(marking.getState(), incidenceMatrix, fMatrix, bMatrix, capacityVector);

		for (int i = 0; i < enabledTransitions.length; i++) {
			if (enabledTransitions[i]) { // If the transition is enabled
//				if (trans.get(i).isTimed()) {
//					// If any immediate transtions exist, the state is vanishing
//					// as they will fire immediately
//					hasTimed = true;
//				} else if (!trans.get(i).isTimed()) {
					hasImmediate = true;
					// return false;
//				}
			}
		}
		return (hasTimed && !hasImmediate);
	}

	/**
	 * fire() Determines all the states resulting from firing enabled
	 * transitions in the state passed as an argument
	 * 
	 * @param pnmlData
     * @param vs
	 *            The state to determine successors from
	 * @param succ
	 *            A stack in which to store successors
     * @return
	 */
	private static int fire(State vs, Stack succ) {
//		int transCount = pn.getAllTransitions().size();

		int transitionsfired = 0;
		double[] newstate = null;
		boolean[] enabledTransitions = PNHelper.getEnabledTransitionsFromMarking(vs.getState(), incidenceMatrix, fMatrix, bMatrix, capacityVector);

		for (int index = 0; index < enabledTransitions.length; index++) {
			if (enabledTransitions[index]) {
				// If the current transition is enabled
				newstate = fireTransition(vs.getState(), index);
				// System.out.println("fired transition " + index);//debug
				succ.push(new State(newstate));
				// System.out.println("generat estat: " + (new
				// State(newstate).toString()));//debug
				transitionsfired++;
				transitions.push(new Integer(index));
			}
		}
		return transitionsfired;
	}

	/**
	 * fireFirstEnabledTransition() Determines the state resulting from firing
	 * enabled transitions in the state passed as an argument
	 * 
	 * @param pnmlData
     * @param enabledTransitions
     * @param vs
	 *            The state to determine successors from
	 * @param succ
	 *            A stack in which to store successors
     * @return
	 */
	private static int fireFirstEnabledTransition(boolean[] enabledTransitions, State vs, Stack succ) {

		for (int index = 0; index < enabledTransitions.length; index++) {
			if (enabledTransitions[index]) {
				// If the current transition is enabled
				double[] newstate = fireTransition(vs.getState(), index);
				// System.out.println("fired transition " + index);//debug
				succ.push(new State(newstate));
				// System.out.println("generat estat: " + (new
				// State(newstate).toString()));//debug
				transitions.push(new Integer(index));
				enabledTransitions[index] = false;
				return index;
			}
		}
		return -1;
	}

	
	
	/**
	 * fireTransition() Produces a new markup vector to simulate the firing of a
	 * transition. Destroys the number of tokens shown in CMinus for a given
	 * place and transition, and creates the number of tokens shown in CPlus.
	 * 
	 * @author Matthew Cook, James Bloom and Clare Clark (original code) Nadeem
	 *         Akharware (optimisation)
	 * @param pnmlData
	 *            The petri net data model
	 * @param marking
	 *            The state/marking to fire from
	 * @param transIndex
	 *            Which transition to fire
	 * @return The new marking/state vector resulting from the fired transition
	 */
	private static double[] fireTransition(double[] marking, int transIndex) {		
		// Create marking array to return
		double[] newmarking = new double[marking.length];
		
		for (int count = 0; count < marking.length; count++) {			
//			newmarking[count] = mo.getIncidenceMatrix().get(count, transIndex) + marking[count];
			newmarking[count] = incidenceMatrix.get(count, transIndex) + marking[count];
		}
		
		return newmarking;		
	}

	/**
	 * explored() Tests whether the state passed as an argument has already been
	 * explored.
	 * 
	 * @param test
	 *            The state to look for
	 * @param es
     * @param exploredStates
	 *            The hashtable to check for the state
	 * @return
	 */
	private static boolean explored(State test, LinkedList[] es) {
		LinkedList hashrow = es[test.hashCode() % NUMHASHROWS];

		if (hashrow == null) {
			// This row has nothing in it yet so must be an unexplored state
			return false;
		}
		Iterator iterator = hashrow.iterator();
		CompressedState current;
        for(Object aHashrow : hashrow)
        {
            current = (CompressedState) iterator.next();
            if(test.hashCode2() == current.getHashCode2())
            {
                return true;
            }
        }
		return false;
	}

	/**
	 * identifyState() Takes a state that we know has been explored before and
	 * works out what id number that state has been given using the explored
	 * states hashtable.
	 * 
	 * @param test
	 *            The state to be identified
	 * @param es
	 *            The hashtable to look it up in
	 * @return The id number of that state (-1 indicates an error)
	 */
	private static int identifyState(State test, LinkedList[] es) {
		LinkedList hashrow = es[test.hashCode() % NUMHASHROWS];
		Iterator iterator = hashrow.iterator();
		CompressedState current;

        for(Object aHashrow : hashrow)
        {
            current = (CompressedState) iterator.next();
            if(test.hashCode2() == current.getHashCode2())
            {
                return current.getID();
            }
        }
		return -1;
	}

	/**
	 * addExplored() Adds a compressed version of a tangible state to the
	 * explored states hashtable and also writes the full state to a file for
	 * later use.
	 * 
	 * @param newstate
	 *            The explored state to be added
	 * @param es
	 *            A reference to the hashtable
	 * @param opfile
	 *            The file to write the state to
     * @param vanishingStates
	 */
	private static void addExplored(MarkingState newstate, LinkedList[] es,
			RandomAccessFile opfile, boolean vanishingStates) {
		LinkedList hashrow = es[newstate.hashCode() % NUMHASHROWS];

		if (hashrow == null) {
			// This hashcode hasn't come up before so we need to set up the
			// linked list first
			es[newstate.hashCode() % NUMHASHROWS] = new LinkedList();
			hashrow = es[newstate.hashCode() % NUMHASHROWS];
		}
		hashrow.add(new CompressedState(newstate.hashCode2(), newstate
				.getIDNum()));
		// Now also write this state to disk for later use
		StateRecord sr = new StateRecord(newstate);
		// System.out.println("afegit: " + sr.toString());//debug
		try {
			if (vanishingStates) {
				sr.write(opfile, newstate.getIsTangible());
			} else {
				sr.write(opfile);
			}
		} catch (IOException e) {
			System.err
					.println("IO problem while writing explored states to file.");
		}
	}

	/**
	 * rate() Calculate the RATE of transition from a TANGIBLE state to another
	 * state. Works out the transitions enabled to fire at a particular marking,
	 * transitions that can be reached from a particular marking and the
	 * intersection of the two. Then sums the firing rates of the intersection
	 * and divides it by the sum of the firing rates of the enabled transitions.
	 * 
	 * @author Matthew Cook (original code), Nadeem Akharware (adaption and
	 *         optimisation), Pere Bonet (minor changes)
	 * @param pnmlData
	 * @param s
     * @param sprime
     * @param v
	 * @param vprime
	 * @return double - the probability
	 */
	private static double rate(State s, State sprime) {
		double[] marking1 = s.getState();
		double[] marking2 = sprime.getState();
		int markSize = marking1.length;
		
		double[][] incidenceMatrix = StateSpaceGenerator.incidenceMatrix.getArrayCopy();
//		int transCount = pn.getAllTransitions().size();

		boolean[] marking1EnabledTransitions = PNHelper.getEnabledTransitionsFromMarking(marking1, StateSpaceGenerator.incidenceMatrix, fMatrix, bMatrix, capacityVector);
		boolean[] matchingTransition = new boolean[marking1EnabledTransitions.length];

		// get transition needed to fire to get from marking1 to marking2
		for (int i = 0; i < marking1EnabledTransitions.length; i++) {
			// initialise value of potential transition to true
			matchingTransition[i] = true;
			for (int k = 0; k < markSize; k++) {
				// if the sum of the incidence matrix and marking 1 doesn't
				// equal
				// marking 2, set that candidate transition possibility to be
				// false
				if ((marking1[k] + incidenceMatrix[k][i]) != marking2[k]) {
					matchingTransition[i] = false;
					break;
				}
			}
		}

		// check if there are any potential transitions from marking 1 to
		// marking
		// 2 and whether they are enabled or not.
		boolean enabledAndMatching = false;
		for (int i = 0; i < marking1EnabledTransitions.length; i++) {
			if (matchingTransition[i]) {
				if (marking1EnabledTransitions[i]) {
					enabledAndMatching = true;
					break; // we have found an enabled and matching transition
				}
			}
		}
		if (!enabledAndMatching) {
			return 0.0;
		}

		// work out the sum of firing weights of input transitions
		double candidateTransitionWeighting = 0.0;
		for (int i = 0; i < marking1EnabledTransitions.length; i++) {
			if ((matchingTransition[i])
					&& (marking1EnabledTransitions[i])) {
//				candidateTransitionWeighting += pn.getAllTransitions().get(i)
//						.getRate();
				candidateTransitionWeighting += 1.0;
			}
		}
		return candidateTransitionWeighting;
	}

	/**
	 * prob() Calculate the PROBABILITY of a transition from a VANISHING state
	 * to another state. Works out the transitions enabled to fire at a
	 * particular marking, transitions that can be reached from a particular
	 * marking and the intersection of the two. Then sums the firing rates of
	 * the intersection and divides it by the sum of the firing rates of the
	 * enabled transitions.
	 * 
	 * @author Matthew Cook (original code), Nadeem Akharware (adaption and
	 *         optimisation), Pere Bonet (minor changes)
	 * @param pnmlData
	 * @param s
     * @param sprime
     * @param firedTransition
     * @param v
	 * @param vprime
	 * @return double - the probability
	 */
	private static double prob(State s, State sprime, int firedTransition) {
		double[] marking1 = s.getState();
		double[] marking2 = sprime.getState();
		int markSize = marking1.length;
//		MatrixOperation mo = new MatrixOperation(pn);
		double[][] incidenceMatrix = StateSpaceGenerator.incidenceMatrix.getArrayCopy();
//		int transCount = pn.getAllTransitions().size();
		boolean[] marking1EnabledTransitions = PNHelper.getEnabledTransitionsFromMarking(marking1, StateSpaceGenerator.incidenceMatrix, fMatrix, bMatrix, capacityVector);
		boolean[] matchingTransition = new boolean[marking1EnabledTransitions.length];

		// get transition needed to fire to get from marking1 to marking2
		for (int i = 0; i < marking1EnabledTransitions.length; i++) {
			// initialise value of potential transition to true
			matchingTransition[i] = true;
			for (int k = 0; k < markSize; k++) {
				// if the sum of the incidence matrix and marking 1 doesn't
				// equal
				// marking 2, set that candidate transition possibility to be
				// false
				if ((marking1[k] + incidenceMatrix[k][i]) != marking2[k]) {
					matchingTransition[i] = false;
					break;
				}
			}
		}

		// check if there are any potential transitions from marking 1 to
		// marking
		// 2 and whether they are enabled or not.
		boolean enabledAndMatching = false;
		for (int i = 0; i < marking1EnabledTransitions.length; i++) {
			if (matchingTransition[i]) {
				if (marking1EnabledTransitions[i]) {
					enabledAndMatching = true;
					break; // we have found an enabled and matching transition
				}
			}
		}
		if (!enabledAndMatching) {
			return 0.0;
		}

		// work out the sum of firing weights of enabled transitions
		double enabledTransitionWeighting = 0.0;
		for (int i = 0; i < marking1EnabledTransitions.length; i++) {
			if (marking1EnabledTransitions[i]) {
//				enabledTransitionWeighting += pn.getAllTransitions().get(i)
//						.getRate();
				enabledTransitionWeighting += 1.0;
			}
		}
//		double candidateTransitionWeighting = pn.getAllTransitions().get(firedTransition)
//				.getRate();
		double candidateTransitionWeighting = 1.0;

		return (candidateTransitionWeighting / enabledTransitionWeighting);
	}

	/**
	 * Records the fact that there is a transition firing sequence from whatever
	 * the current tangible state is to the tangible state sprime with an
	 * effective transition firing rate r. Note it does not need to know what
	 * the current tangible state actually is, it just needs a reference to the
	 * list of arcs from that state.
	 * 
	 * @param sprime
	 * @param r
	 * @param arclist
	 *            A linked list of arcs from the current tangible state.
     * @return
	 */
	private static int transition(MarkingState sprime, double r, LinkedList arclist) {
		ArcListElement current;
		if (arclist.size() > 0) {
			Iterator iterator = arclist.iterator();
			current = (ArcListElement) iterator.next();
			while ((current.getTo() != sprime.getIDNum()) && iterator.hasNext()) {
				current = (ArcListElement) iterator.next();
			}
			if (current.getTo() == sprime.getIDNum()) {
				double rate = current.getRate();
				current.setRate(r + rate);
				return 0;
			} else {
				current = new ArcListElement(sprime.getIDNum(), r,
						(Integer) transitions.pop());
				arclist.add(current);
				return 1;
			}
		} else {
			// This must be a new arc
			current = new ArcListElement(sprime.getIDNum(), r,
					(Integer) transitions.pop());
			arclist.add(current);
			return 1;
		}
	}
	
	/**
	 * writeTransitions() Records all the arcs in the reachability graph from
	 * state 'from'.
	 * 
	 * @param from
	 *            The tangible state which all the arcs in the linked list are
	 *            from.
	 * @param arclist
	 *            The list of arcs.
	 * @param dataFile
	 *            The file that reachability graph data needs to be written to.
	 * @param writeTransitionsNo
	 * @throws ImmediateAbortException
	 */
	private static void writeTransitions(MarkingState from, LinkedList arclist,
			RandomAccessFile dataFile, boolean writeTransitionsNo)
			throws ImmediateAbortException {		
		TransitionRecord newTransition;
		Iterator iterator = arclist.iterator();
		ArcListElement current;

		while (iterator.hasNext()) {
			current = (ArcListElement) iterator.next();
			if (writeTransitionsNo) {
				newTransition = new TransitionRecord(from.getIDNum(), current
						.getTo(), current.getRate(), current.transitionNo);
			
				try {
					newTransition.write1(dataFile);
					// System.out.println("From: " + from.getIDNum() + ":(" +
					// from + ")" + " To: " + current.getTo() + " via " +
					// current.transitionNo+ " Rate: " + current.getRate());
				} catch (IOException e) {
					System.err
							.println("IO error when writing transitions to file.");
					throw new ImmediateAbortException();
				}
			} else {
				newTransition = new TransitionRecord(from.getIDNum(), current
						.getTo(), current.getRate());
				try {
					newTransition.write(dataFile);
					// System.out.println("From: " + from.getIDNum() + ":(" +
					// from + ")" + " To: " + current.getTo() + " Rate: " +
					// current.getRate());
				} catch (IOException e) {
					System.err
							.println("IO error when writing transitions to file.");
					throw new ImmediateAbortException();
				}
			}
		}
	}
	
	/**
	 * createRGFile() Creates a reachability graph file containing all the
	 * tangible states that were found during state space exploration and also
	 * all the transitions between them.
	 * 
	 * @param transource
	 *            A file containing all the transitions (arcs) between tangible
	 *            states.
	 * @param destination
	 *            The file to create as a reachability graph file. The file
	 *            should already contain a list of all the tangible states and
	 *            be in the correct position for writing the record of all the
	 *            transitions between them.
	 * @param statesize
	 *            The size of each state array
	 * @param states
	 *            The number of tangible states found
	 * @param transitions
	 *            The number of transitions recorded
     * @param withTransitions
	 * @throws IOException 
	 */
	private static void createRGFile(File transource,
			RandomAccessFile destination, int statesize, int states,
			int transitions, boolean withTransitions) throws IOException {
		RandomAccessFile transinputFile = null;
		TransitionRecord currenttran = new TransitionRecord();
		ReachabilityGraphFileHeader header;
//		try {
			transinputFile = new RandomAccessFile(transource, "r");

			// The destination file actually already exists with a blank file
			// header as a placeholder and all the tangible states written in
			// order. The file pointer should already be at the end of the file
			// (i.e. after the last tangible state that's been written to the
			// file.
			// Make a note of the file pointer as this is where the transition
			// records begin.
			long offset = destination.getFilePointer();
			// Now copy over all the transitions
			System.out.println("Creating reachability graph, please wait...");
			for (int count = 0; count < transitions; count++) {
				// System.out.print("Recording arc " + (count+1) + " of " +
				// transitions +".\r");
				if (withTransitions) {
					currenttran.read1(transinputFile);
					currenttran.write1(destination);
				} else {
					currenttran.read(transinputFile);
					currenttran.write(destination);
				}
			}

			// Make a note of the transition record size and fill in all the
			// details in the file header.
			int recordsize = currenttran.getRecordSize();
			destination.seek(0); // Go back to the start of the file
			header = new ReachabilityGraphFileHeader(states, statesize, transitions,
					recordsize, offset);
			header.write(destination);

			// Done so close all the files.
			transinputFile.close();
//			destination.close();
//		} catch (EOFException e) {
//			System.err.println("EOFException");
//		} catch (IOException e) {
//			System.out.println("Could not create output file.");
//			e.getMessage();
//		}
//		finally{
//			try {
//				transinputFile.close();
//				destination.close();
//			} catch (IOException e) {
//				// TODO Auto-generated catch block
//				e.printStackTrace();
//			}			
//		}
	}

	private static void printArray(boolean[] array) {
		int rows = array.length;
		System.out.print("Elements as follows: ");
		for (int i = 0; i < rows; i++) {
			System.out.print(array[i] + " ");
		}
		System.out.println();
	}

	private static void printArray(double[] array) {
		int rows = array.length;
		System.out.print("Elements as follows: ");
		for (int i = 0; i < rows; i++) {
			System.out.print(array[i] + " ");
		}
		System.out.println();
	}

	private static void printArray(State state) {
		double[] array = state.getState();
		printArray(array);
	}	
}


